<?php
/**
 * BSS Commerce Co.
 *
 * NOTICE OF LICENSE
 *
 * @category   BSS
 * @package    Bss_InventoryReport
 * @author     Extension Team
 * @copyright  Copyright (c) 2018-2019 BSS Commerce Co. ( http://bsscommerce.com )
 * @license    http://bsscommerce.com/Bss-Commerce-License.txt
 */
namespace Bss\InventoryReport\Model;

use Bss\InventoryReport\Api\Data\InventoryInterface;
use Bss\InventoryReport\Api\InventoryRepositoryInterface;
use Magento\Framework\Api\SearchCriteriaInterface;
use Bss\InventoryReport\Api\Data\InventorySearchResultsInterfaceFactory;
use Magento\Framework\Exception\CouldNotDeleteException;
use Magento\Framework\Exception\CouldNotSaveException;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Framework\Exception\StateException;

class InventoryRepository implements InventoryRepositoryInterface
{
    /**
     * @var \Bss\InventoryReport\Model\ReportFactory
     */
    protected $reportFactory;

    /**
     * @var ResourceModel\Report
     */
    protected $reportResource;

    /**
     * @var ResourceModel\Report\CollectionFactory
     */
    protected $collectionFactory;

    /**
     * @var InventorySearchResultsInterfaceFactory
     */
    protected $searchResultsFactory;

    /**
     * InventoryRepository constructor.
     * @param ReportFactory $reportFactory
     * @param ResourceModel\Report $reportResource
     * @param ResourceModel\Report\CollectionFactory $collectionFactory
     * @param InventorySearchResultsInterfaceFactory $searchResultsFactory
     */
    public function __construct(
        \Bss\InventoryReport\Model\ReportFactory $reportFactory,
        \Bss\InventoryReport\Model\ResourceModel\Report $reportResource,
        \Bss\InventoryReport\Model\ResourceModel\Report\CollectionFactory $collectionFactory,
        InventorySearchResultsInterfaceFactory $searchResultsFactory
    ) {
        $this->reportFactory = $reportFactory;
        $this->reportResource = $reportResource;
        $this->collectionFactory = $collectionFactory;
        $this->searchResultsFactory = $searchResultsFactory;
    }

    /**
     * Get Report by id
     *
     * @param int $reportId
     * @return InventoryInterface|void
     * @throws NoSuchEntityException
     */
    public function getById($reportId)
    {
        $report = $this->reportFactory->create();
        $this->reportResource->load($report, $reportId);
        if (!$report->getReportId()) {
            throw new NoSuchEntityException(
                __(
                    'The report with the "%1" ID wasn\'t found. Verify the ID and try again.',
                    $report->getReportId()
                )
            );
        }
        return $report;
    }

    /**
     * Delete report by id
     *
     * @param int $reportId
     * @return bool
     * @throws CouldNotDeleteException
     * @throws NoSuchEntityException
     */
    public function deleteById($reportId)
    {
        $report = $this->getById($reportId);
        try {
            $this->delete($report);
        } catch (\Exception $exception) {
            throw new CouldNotDeleteException(
                __('Cannot delete report with id %1', $reportId),
                $exception
            );
        }
        return true;
    }

    /**
     * Save report
     *
     * @param InventoryInterface $inventory
     * @return InventoryInterface
     * @throws CouldNotSaveException
     */
    public function save(InventoryInterface $inventory)
    {
        try {
            $this->reportResource->save($inventory);
        } catch (\Exception $e) {
            throw new CouldNotSaveException(
                __(
                    'Could not save report: %1',
                    $e->getMessage()
                ),
                $e
            );
        }
        return $inventory;
    }

    /**
     * Delete report
     *
     * @param InventoryInterface $inventory
     * @return bool
     * @throws StateException
     */
    public function delete(InventoryInterface $inventory)
    {
        try {
            $this->reportResource->delete($inventory);
        } catch (\Exception $e) {
            throw new StateException(
                __(
                    'Cannot delete report with id %1',
                    $inventory->getReportId()
                ),
                $e
            );
        }
        return true;
    }

    /**
     * Get list items
     *
     * @param SearchCriteriaInterface $searchCriteria
     * @return \Bss\InventoryReport\Api\Data\InventorySearchResultsInterface
     */
    public function getList(SearchCriteriaInterface $searchCriteria)
    {
        $collection = $this->collectionFactory->create();
        $collection->load();
        $searchResult = $this->searchResultsFactory->create();
        $searchResult->setSearchCriteria($searchCriteria);
        $searchResult->setItems($collection->getItems());
        $searchResult->setTotalCount($collection->getSize());
        return $searchResult;
    }
}
