<?php
/**
 * BSS Commerce Co.
 *
 * NOTICE OF LICENSE
 *
 * @category   BSS
 * @package    Bss_InventoryReport
 * @author     Extension Team
 * @copyright  Copyright (c) 2018-2019 BSS Commerce Co. ( http://bsscommerce.com )
 * @license    http://bsscommerce.com/Bss-Commerce-License.txt
 */
namespace Bss\InventoryReport\Model\MultiInventorySource;

use Magento\Framework\Api\SearchCriteriaBuilder;

class StockAndSourceProcessor
{
    /**
     * Default sales channel type
     */
    const SALES_CHANNEL_TYPE_WEBSITE = 'website';
    const SOURCE_ITEM_SKU = 'sku';
    const SOURCE_ITEM_SOURCE_CODE = 'source_code';

    /**
     * @var MsiObjectManagement
     */
    protected $msiObjectManagement;

    /**
     * @var SearchCriteriaBuilder
     */
    protected $searchCriteriaBuilder;

    /**
     * StockAndSourceProcessor constructor.
     * @param MsiObjectManagement $msiObjectManagement
     * @param SearchCriteriaBuilder $searchCriteriaBuilder
     */
    public function __construct(
        MsiObjectManagement $msiObjectManagement,
        SearchCriteriaBuilder $searchCriteriaBuilder
    ) {
        $this->msiObjectManagement = $msiObjectManagement;
        $this->searchCriteriaBuilder = $searchCriteriaBuilder;
    }

    /**
     * Get Source Item From Product Sku And Source Code
     *
     * Used For MSI only
     *
     * @param string $productSku
     * @param string $sourceCode
     * @return null|array|mixed
     */
    public function getSourceItem($productSku, $sourceCode)
    {
        $getSourceItemRepository = $this->msiObjectManagement->getSourceItemRepository();
        if ($getSourceItemRepository !== null &&
            $getSourceItemRepository instanceof \Magento\InventoryApi\Api\SourceItemRepositoryInterface) {

            $searchCriteria = $this->searchCriteriaBuilder
                ->addFilter(self::SOURCE_ITEM_SKU, $productSku)
                ->addFilter(self::SOURCE_ITEM_SOURCE_CODE, $sourceCode)
                ->create();
            $sourceItems = $getSourceItemRepository->getList($searchCriteria)->getItems();
            return array_shift($sourceItems);
        }
        return null;
    }

    /**
     * Get List Source Item From Product Sku
     *
     * Used For MSI only
     *
     * @param string $productSku
     * @return null|object|mixed
     */
    public function getListSource($productSku)
    {
        $getSourceItemsBySku = $this->msiObjectManagement->getSourceItemsBySku();
        if ($getSourceItemsBySku !== null &&
            $getSourceItemsBySku instanceof \Magento\InventoryApi\Api\GetSourceItemsBySkuInterface) {
            return $getSourceItemsBySku->execute($productSku);
        }
        return null;
    }

    /**
     * Get Revervation qty order
     *
     * @param string $sku
     * @param int $stockId
     * @return null|float
     */
    public function getReservationQty($sku, $stockId)
    {
        $object = $this->msiObjectManagement->getReservationsQuantity();
        if ($object !== null &&
            $object instanceof \Magento\InventoryReservationsApi\Model\GetReservationsQuantityInterface) {
            return $object->execute($sku, $stockId);
        }
        return null;
    }

    /**
     * Get saleable Qty by sku
     *
     * @param string $productSku
     * @return null|array
     */
    public function getSaleableQty($productSku)
    {
        $getSalableQuantityDataBySku = $this->msiObjectManagement->getSalableQuantityDataBySku();
        if ($getSalableQuantityDataBySku !== null &&
            $getSalableQuantityDataBySku instanceof \Magento\InventorySalesAdminUi\Model\GetSalableQuantityDataBySku) {
            return $getSalableQuantityDataBySku->execute($productSku);
        }
        return null;
    }

    /**
     * Get stock by source code
     *
     * @param string $sourceCode
     * @return null|array|object[]|mixed
     */
    public function getStockBySourceCode($sourceCode)
    {
        $object = $this->msiObjectManagement->getStockSourceLinksBySourceCode();
        if ($object !== null &&
            $object instanceof \Magento\InventorySalesAdminUi\Model\GetStockSourceLinksBySourceCode) {
            return $object->execute($sourceCode);
        }
        return null;
    }

    /**
     * Get saleable qty by product sku and stock id
     *
     * @param string $productSku
     * @param int $stockId
     * @return null|float
     * @throws \Magento\Framework\Exception\InputException
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function getSaleableQtyByStockId($productSku, $stockId)
    {
        $getProductSalableQty = $this->msiObjectManagement->getProductSalableQty();
        if ($getProductSalableQty !== null &&
            $getProductSalableQty instanceof \Magento\InventorySalesApi\Api\GetProductSalableQtyInterface) {
            return $getProductSalableQty->execute($productSku, $stockId);
        }
        return null;
    }

    /**
     * Get source code by stock id
     *
     * @param int  $stockId
     * @return bool|string|null
     */
    public function getSourceCodeByStockId($stockId)
    {
        try {
            $getStockSourceLinkFactory = $this->msiObjectManagement->getStockSourceLinkFactory();
            if ($getStockSourceLinkFactory !== null &&
                $getStockSourceLinkFactory instanceof \Magento\Inventory\Model\StockSourceLinkFactory) {
                $stockLink = $getStockSourceLinkFactory->create()->load($stockId, 'stock_id');
                return $stockLink->getSourceCode();
            }
        } catch (\Exception $e) {
            return false;
        }
        return false;
    }

    /**
     * Get stock by webiste code
     *
     * @param string $websiteCode
     * @return null|object|mixed
     * @throws \Magento\Framework\Exception\NoSuchEntityException
     */
    public function getStockByWebsiteCode($websiteCode)
    {
        $getStockResolver = $this->msiObjectManagement->getStockResolver();
        if ($getStockResolver !== null &&
            $getStockResolver instanceof \Magento\InventorySalesApi\Api\StockResolverInterface) {
            return $getStockResolver->execute(self::SALES_CHANNEL_TYPE_WEBSITE, $websiteCode);
        }
        return null;
    }

    /**
     *
     * @param int $stockId
     * @return bool|string|null|object
     */
    public function getSourceProductByStockId($stockId)
    {
        try {
            $getStockSourceLinkFactory = $this->msiObjectManagement->getStockSourceLinkFactory();
            if ($getStockSourceLinkFactory !== null &&
                $getStockSourceLinkFactory instanceof \Magento\Inventory\Model\StockSourceLinkFactory) {
                $stockLinkCollection = $getStockSourceLinkFactory->create()
                    ->getCollection()
                    ->addFieldToFilter('stock_id', $stockId);

                return $stockLinkCollection;
            }
        } catch (\Exception $e) {
            return false;
        }
        return false;
    }
}
