<?php
/**
 * BSS Commerce Co.
 *
 * NOTICE OF LICENSE
 *
 * @category   BSS
 * @package    Bss_InventoryReport
 * @author     Extension Team
 * @copyright  Copyright (c) 2018-2019 BSS Commerce Co. ( http://bsscommerce.com )
 * @license    http://bsscommerce.com/Bss-Commerce-License.txt
 */

namespace Bss\InventoryReport\Observer;

use Bss\InventoryReport\Helper\InventoryData as InventoryDataHelper;
use Exception;
use Magento\Catalog\Model\Product;
use Magento\Catalog\Model\ProductFactory;
use Magento\Catalog\Model\ResourceModel\Product\Collection;
use Magento\Catalog\Model\ResourceModel\Product\CollectionFactory as ProductCollectionFactory;
use Magento\CatalogInventory\Api\StockRegistryInterface;
use Magento\Framework\App\RequestInterface;
use Magento\Framework\Event\Observer;
use Magento\Framework\Event\ObserverInterface;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Framework\Message\ManagerInterface;
use Magento\Framework\Registry;
use Magento\Framework\Stdlib\DateTime\DateTime;

class EditorInlineSaveBefore implements ObserverInterface
{
    /**
     * Const
     */
    const DEFAULT_SOURCE = 'default';
    const KEY_DATA_INLINE_EDIT_SAVE_BEFORE = 'bss_data_inline_edit_save_before';

    /**
     * @var RequestInterface
     */
    protected $request;

    /**
     * @var Registry
     */
    protected $coreRegistry;

    /**
     * @var ProductFactory
     */
    protected $productFactory;

    /**
     * @var ProductCollectionFactory
     */
    protected $productCollectionFactory;

    /**
     * @var ManagerInterface
     */
    protected $messageManager;

    /**
     * @var DateTime
     */
    protected $dateTime;

    /**
     * @var StockRegistryInterface
     */
    protected $stockRegistry;

    /**
     * @var InventoryDataHelper
     */
    protected $inventoryDataHelper;

    /**
     * EditorInlineSaveBefore constructor.
     * @param RequestInterface $request
     * @param Registry $coreRegistry
     * @param ProductFactory $productFactory
     * @param ProductCollectionFactory $productCollectionFactory
     * @param ManagerInterface $messageManager
     * @param DateTime $dateTime
     * @param StockRegistryInterface $stockRegistry
     * @param InventoryDataHelper $inventoryDataHelper
     */
    public function __construct(
        RequestInterface $request,
        Registry $coreRegistry,
        ProductFactory $productFactory,
        ProductCollectionFactory $productCollectionFactory,
        ManagerInterface $messageManager,
        DateTime $dateTime,
        StockRegistryInterface $stockRegistry,
        InventoryDataHelper $inventoryDataHelper
    ) {
        $this->request = $request;
        $this->coreRegistry = $coreRegistry;
        $this->productFactory = $productFactory;
        $this->productCollectionFactory = $productCollectionFactory;
        $this->messageManager = $messageManager;
        $this->dateTime = $dateTime;
        $this->stockRegistry = $stockRegistry;
        $this->inventoryDataHelper = $inventoryDataHelper;
    }

    /**
     * @param Observer $observer
     * @return $this|void
     */
    public function execute(Observer $observer)
    {
        try {
            $productIds = $observer->getData('products');
            /** @var Collection $productCollection */
            $productCollection = $this->productCollectionFactory->create();

            $productCollection->addAttributeToFilter('entity_id', ['in' => $productIds])
                ->addAttributeToSelect(['name', 'sku', 'entity_id']);
            $this->regisProductInventoryReportData($productCollection);
        } catch (Exception $exception) {
            $this->messageManager->addWarningMessage(
                __('Stock Report did not update due unexpected error. Please see log to more detail.')
            );
        }
        return $this;
    }

    /**
     * @param Collection $productItems
     * @throws NoSuchEntityException
     */
    protected function regisProductInventoryReportData($productItems)
    {
        $dataRenderer = [];
        /** @var Product $productItem */
        foreach ($productItems as $productItem) {
            if ($productItem->getSku()) {
                $qtyBefore = (float)$this->getQtyProductItem($productItem->getSku());
                $qtyAfter = (float)$this->getQtyProductItemAfter($productItem->getSku());
                $qtyChange = $qtyAfter - $qtyBefore;

                $data = $this->createBundleData($productItem);
                $data['source_code'] = self::DEFAULT_SOURCE;
                $data['qty_after'] = $qtyAfter;
                $data['qty_change'] = $qtyChange;
                if ($qtyChange) {
                    $dataRenderer[$productItem->getId()] = $data;
                }
            }
        }
        $this->coreRegistry->register(self::KEY_DATA_INLINE_EDIT_SAVE_BEFORE, $dataRenderer);
    }

    /**
     * @param string $productItemSku
     * @return float|int
     * @throws NoSuchEntityException
     */
    protected function getQtyProductItem($productItemSku)
    {
        // Before save, qty can be get by stock
        $productStock = $this->stockRegistry->getStockItemBySku($productItemSku);
        if ($productStock->getQty()) {
            return $productStock->getQty();
        }
        return 0;
    }

    /**
     * @param string $productItemSku
     * @return int|mixed
     */
    protected function getQtyProductItemAfter($productItemSku)
    {
        // Get qty is preparing to save by post params
        $postItems = $this->request->getParam('items', []);
        $qtyPost = $this->mixQtyPostToArray($postItems);
        if (isset($qtyPost[$productItemSku])) {
            return $qtyPost[$productItemSku];
        }
        return 0;
    }

    /**
     * @param $postItems
     * @return array
     */
    protected function mixQtyPostToArray($postItems)
    {
        $qtyPost = [];
        foreach ($postItems as $postItem) {
            if (isset($postItem['sku']) &&
                isset($postItem['qty'])) {
                $qtyPost[$postItem['sku']] = $postItem['qty'];
            }
        }
        return $qtyPost;
    }

    /**
     * @param Product $product
     * @return array
     */
    protected function createBundleData($product)
    {
        return [
            'time' => $this->dateTime->gmtDate(),
            'user' => $this->inventoryDataHelper->getAdminName(),
            'note' => __('Admin Manual Change'),
            'product_id' => $product->getId(),
            'product_name' => $product->getName(),
            'product_sku' => $product->getSku()
        ];
    }
}
