<?php
/**
 * BSS Commerce Co.
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the EULA
 * that is bundled with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://bsscommerce.com/Bss-Commerce-License.txt
 *
 * @category   BSS
 * @package    Bss_InventoryReport
 * @author     Extension Team
 * @copyright  Copyright (c) 2017-2018 BSS Commerce Co. ( http://bsscommerce.com )
 * @license    http://bsscommerce.com/Bss-Commerce-License.txt
 */

namespace Bss\InventoryReport\Observer;

use Bss\InventoryReport\Helper\InventoryData as InventoryDataHelper;
use Bss\InventoryReport\Model\MultiInventorySource\MsiObjectManagement;
use Bss\InventoryReport\Model\ResourceModel\Report as InventoryResourceReport;
use Exception;
use Magento\Catalog\Model\Product;
use Magento\Catalog\Model\ProductRepository;
use Magento\CatalogInventory\Api\Data\StockItemInterface;
use Magento\CatalogInventory\Api\StockRegistryInterface;
use Magento\Framework\Event\Observer;
use Magento\Framework\Event\ObserverInterface;
use Magento\Framework\Stdlib\DateTime\DateTime;
use Psr\Log\LoggerInterface;

class ImportObserver implements ObserverInterface
{
    /**
     * @var DateTime
     */
    protected $date;

    /**
     * @var LoggerInterface
     */
    protected $logger;

    /**
     * @var ProductRepository
     */
    protected $productRepository;

    /**
     * @var MsiObjectManagement
     */
    protected $msiObjectManagement;

    /**
     * @var StockRegistryInterface
     */
    protected $stockRegistry;

    /**
     * @var InventoryDataHelper
     */
    protected $inventoryDataHelper;

    /**
     * @var InventoryResourceReport
     */
    protected $inventoryResourceReport;

    /**
     * ImportObserver constructor.
     * @param DateTime $date
     * @param StockRegistryInterface $stockRegistry
     * @param InventoryResourceReport $inventoryResourceReport
     * @param ProductRepository $productRepository
     * @param LoggerInterface $logger
     * @param MsiObjectManagement $msiObjectManagement
     * @param InventoryDataHelper $inventoryDataHelper
     */
    public function __construct(
        DateTime $date,
        LoggerInterface $logger,
        StockRegistryInterface $stockRegistry,
        ProductRepository $productRepository,
        MsiObjectManagement $msiObjectManagement,
        InventoryDataHelper $inventoryDataHelper,
        InventoryResourceReport $inventoryResourceReport
    ) {
        $this->date = $date;
        $this->logger = $logger;
        $this->stockRegistry = $stockRegistry;
        $this->productRepository = $productRepository;
        $this->msiObjectManagement = $msiObjectManagement;
        $this->inventoryDataHelper = $inventoryDataHelper;
        $this->inventoryResourceReport = $inventoryResourceReport;
    }

    /**
     * @param Observer $observer
     * @return void
     */
    public function execute(Observer $observer)
    {
        // Only process this if MSI isn't enabled
        if (!$this->msiObjectManagement->isMsiEnabled()) {
            try {
                $adapter = $observer->getEvent()->getAdapter();
                $products = $adapter->getNewSku();

                foreach ($products as $product) {
                    if (isset($product['entity_id']) && $product['entity_id']) {
                        $productId = $product['entity_id'];
                        $product = $this->productRepository->getById($productId);

                        if ($this->isAllowProductType($product)) {
                            $stockItem = $this->stockRegistry->getStockItemBySku($product->getSku());
                            $stockQtyAfter = $stockItem->getQty();

                            $stockStatus = $this->getStockStatus($stockItem, $stockQtyAfter);
                            $this->inventoryResourceReport->updateImportProduct(
                                $product->getSku(),
                                $this->date->gmtDate(),
                                $productId,
                                __('Import Product'),
                                $product->getName(),
                                $stockQtyAfter,
                                $stockStatus,
                                $this->inventoryDataHelper->getAdminName()
                            );
                        }
                    }
                }
            } catch (Exception $e) {
                $this->logger->debug($e->getMessage());
            }
        }
    }

    /**
     * @param Product $product
     * @return bool
     */
    protected function isAllowProductType($product)
    {
        $productType = $product->getTypeId();
        return $productType == 'simple' || $productType == 'virtual';
    }

    /**
     * @param StockItemInterface $stockItem
     * @param float|int|string $stockQtyAfter
     * @return string
     */
    protected function getStockStatus($stockItem, $stockQtyAfter)
    {
        $outStockThreshold = $stockItem->getMinQty();
        $backOrderConfigProduct = $stockItem->getBackorders();
        $backOrderValue = $this->inventoryDataHelper->getBackOrderValue();
        $getUseConfigBackOrder = $stockItem->getUseConfigBackorders();
        if (($stockQtyAfter > $outStockThreshold)
            || (($backOrderValue != 0 && $getUseConfigBackOrder)
                || $backOrderConfigProduct != 0)
        ) {
            $stockStatus = 1;
        } else {
            $stockStatus = 0;
        }
        return $stockStatus;
    }
}
