<?php
/**
 * BSS Commerce Co.
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the EULA
 * that is bundled with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://bsscommerce.com/Bss-Commerce-License.txt
 *
 * @category   BSS
 * @package    Bss_InventoryReport
 * @author     Extension Team
 * @copyright  Copyright (c) 2017-2018 BSS Commerce Co. ( http://bsscommerce.com )
 * @license    http://bsscommerce.com/Bss-Commerce-License.txt
 */

namespace Bss\InventoryReport\Observer;

use Bss\InventoryReport\Helper\InventoryData as InventoryDataHelper;
use Bss\InventoryReport\Model\MultiInventorySource\MsiObjectManagement;
use Bss\InventoryReport\Model\MultiInventorySource\StockAndSourceProcessor;
use Bss\InventoryReport\Model\ReportFactory;
use Bss\InventoryReport\Model\ResourceModel\Report as InventoryResourceReport;
use Exception;
use Magento\CatalogInventory\Api\Data\StockItemInterface;
use Magento\CatalogInventory\Api\StockRegistryInterface;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\Event\Observer;
use Magento\Framework\Event\ObserverInterface;
use Magento\Framework\Exception\InputException;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Framework\Stdlib\DateTime\DateTime;
use Magento\Framework\Stdlib\DateTime\TimezoneInterface;
use Magento\Sales\Model\Order\Item as OrderItem;
use Magento\Store\Api\WebsiteRepositoryInterface;
use Psr\Log\LoggerInterface;

class OrderCancelObserver implements ObserverInterface
{
    /**
     * @var WebsiteRepositoryInterface
     */
    protected $websiteRepository;

    /**
     * @var MsiObjectManagement
     */
    protected $msiObjectManagement;

    /**
     * @var StockAndSourceProcessor
     */
    protected $stockAndSourceProcessor;

    /**
     * @var TimezoneInterface
     */
    protected $timezone;

    /**
     * @var ReportFactory
     */
    protected $reportFactory;

    /**
     * @var DateTime
     */
    protected $dateTime;

    /**
     * @var StockRegistryInterface
     */
    protected $stockRegistry;

    /**
     * @var InventoryResourceReport
     */
    protected $inventoryResourceReport;

    /**
     * @var ScopeConfigInterface
     */
    protected $scopeConfig;

    /**
     * @var InventoryDataHelper
     */
    protected $inventoryData;

    /**
     * @var LoggerInterface
     */
    protected $logger;

    /**
     * OrderCancelObserver constructor.
     * @param WebsiteRepositoryInterface $websiteRepository
     * @param MsiObjectManagement $msiObjectManagement
     * @param StockAndSourceProcessor $stockAndSourceProcessor
     * @param TimezoneInterface $timezone
     * @param ReportFactory $reportFactory
     * @param DateTime $dateTime
     * @param StockRegistryInterface $stockRegistry
     * @param InventoryResourceReport $inventoryResourceReport
     * @param ScopeConfigInterface $scopeConfig
     * @param InventoryDataHelper $inventoryData
     * @param LoggerInterface $logger
     */
    public function __construct(
        WebsiteRepositoryInterface $websiteRepository,
        MsiObjectManagement $msiObjectManagement,
        StockAndSourceProcessor $stockAndSourceProcessor,
        TimezoneInterface $timezone,
        ReportFactory $reportFactory,
        DateTime $dateTime,
        StockRegistryInterface $stockRegistry,
        InventoryResourceReport $inventoryResourceReport,
        ScopeConfigInterface $scopeConfig,
        InventoryDataHelper $inventoryData,
        LoggerInterface $logger
    ) {
        $this->websiteRepository = $websiteRepository;
        $this->msiObjectManagement = $msiObjectManagement;
        $this->stockAndSourceProcessor = $stockAndSourceProcessor;
        $this->timezone = $timezone;
        $this->reportFactory = $reportFactory;
        $this->dateTime = $dateTime;
        $this->stockRegistry = $stockRegistry;
        $this->inventoryResourceReport = $inventoryResourceReport;
        $this->scopeConfig = $scopeConfig;
        $this->inventoryData = $inventoryData;
        $this->logger = $logger;
    }

    /**
     * @param Observer $observer
     * @throws InputException
     * @throws LocalizedException
     * @throws NoSuchEntityException
     */
    public function execute(Observer $observer)
    {
        if ($this->msiObjectManagement->isMsiEnabled()) {
            $this->processWithMsi($observer);
        } else {
            $this->processWithoutMsi($observer);
        }
    }

    /**
     * @param ObserverInterface $observer
     * @throws InputException
     * @throws LocalizedException
     * @throws NoSuchEntityException
     */
    protected function processWithMsi($observer)
    {
        $data = [];

        /** @var OrderItem $orderItem */
        $orderItem = $observer->getEvent()->getItem();
        $qtyOrdered = $orderItem->getQtyOrdered();
        $qtyShipped = $orderItem->getQtyShipped();
        $websiteId = $orderItem->getStore()->getWebsiteId();
        $order = $orderItem->getOrder();

        $data['time'] = $this->timezone->date();
        $data['note'] = $this->getNote($order->getIncrementId());
        $data['user'] = $this->inventoryData->getAdminName();

        $websiteCode = $this->websiteRepository->getById($websiteId)->getCode();
        $stockId = $this->stockAndSourceProcessor->getStockByWebsiteCode($websiteCode)->getStockId();
        $stockLinkCollection = $this->stockAndSourceProcessor->getSourceProductByStockId($stockId);

        foreach ($stockLinkCollection as $stockLink) {
            $sourceCode = $stockLink->getSourceCode();
            $data['source_code'] = $sourceCode;
            if ($this->isAllowProductType($orderItem)) {
                $productSku = $orderItem->getSku();
                $sourceItem = $this->stockAndSourceProcessor->getSourceItem($productSku, $sourceCode);
                if ($sourceItem && $sourceItem->getQuantity() > 0) {
                    $salableQty = $this->stockAndSourceProcessor->getSaleableQtyByStockId($productSku, $stockId);
                    $data['stock_id'] = $stockId;
                    $data['product_id'] = $orderItem->getProductId();
                    $data['product_name'] = $orderItem->getName();
                    $data['product_sku'] = $orderItem->getSku();
                    $data['qty_change'] = $qtyShipped;
                    $data['qty_after'] = $sourceItem->getQuantity();
                    $data['stock_status'] = $sourceItem->getStatus();
                    $data['saleable_qty'] = $salableQty;
                    $data['saleable_qty_change'] = $qtyOrdered - $qtyShipped;
                    $data['order_increment_id'] = $order->getIncrementId();
                    $this->reportFactory->create()->setData($data)->save();
                    break;
                }
            }
        }
    }

    /**
     * @param string|int $orderId
     * @return string
     */
    protected function getNote($orderId)
    {
        return __('Order Canceled : %1', $orderId);
    }

    /**
     * @param OrderItem $orderItem
     * @return bool
     */
    protected function isAllowProductType($orderItem)
    {
        $productType = $orderItem->getProductType();
        return $productType == 'simple' || $productType == 'virtual' || $productType == 'grouped';
    }

    /**
     * @param Observer $observer
     */
    protected function processWithoutMsi($observer)
    {
        try {
            $user = $this->inventoryData->getAdminName();
            $currentTime = $this->dateTime->gmtDate();
            /** @var OrderItem $orderItem */
            $orderItem = $observer->getEvent()->getItem();
            $order = $orderItem->getOrder();
            $incrementId = $order->getIncrementId();

            if ($this->isAllowProductType($orderItem)) {
                $productId = $orderItem->getProductId();
                $productName = $orderItem->getName();
                $productSku = $orderItem->getSku();

                $qtyChange = $orderItem->getQtyOrdered();
                $qtyChange = round($qtyChange, 0);
                $stockItem = $this->stockRegistry->getStockItemBySku($productSku);
                $stockQtyAfter = $stockItem->getQty();
                $stockStatus = $this->getStatus($stockItem, $stockQtyAfter, $productSku);
                $note = $this->getNote($incrementId);

                $this->inventoryResourceReport->saveReport(
                    $currentTime,
                    $productId,
                    $productName,
                    $productSku,
                    $stockQtyAfter,
                    $qtyChange,
                    $stockStatus,
                    $user,
                    $note,
                    $orderItem->getProductType(),
                    $incrementId
                );
            }
        } catch (Exception $e) {
            $this->logger->debug($e->getMessage());
        }
    }

    /**
     * @param StockItemInterface $stockItem
     * @param string|int $stockQtyAfter
     * @param string $productSku
     * @return int|string
     */
    protected function getStatus($stockItem, $stockQtyAfter, $productSku)
    {
        $minQtyConfigProduct = $stockItem->getMinQty();
        $useMinQtyConfig = $stockItem->getUseConfigMinQty();

        $threshold = $this->inventoryData->getThresholdValue();
        $stockStatus = $this->stockRegistry->getProductStockStatusBySku($productSku);

        if ($stockStatus == 1 || $stockQtyAfter > 0) {
            if ($useMinQtyConfig && $stockQtyAfter > $threshold) {
                $stockStatus = 1;
            } elseif ($stockQtyAfter > $minQtyConfigProduct) {
                $stockStatus = 1;
            } else {
                $stockStatus = 0;
            }
        } else {
            $stockStatus = 0;
        }
        return $stockStatus;
    }
}
