<?php
/**
 * BSS Commerce Co.
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the EULA
 * that is bundled with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://bsscommerce.com/Bss-Commerce-License.txt
 *
 * @category   BSS
 * @package    Bss_InventoryReport
 * @author     Extension Team
 * @copyright  Copyright (c) 2017-2018 BSS Commerce Co. ( http://bsscommerce.com )
 * @license    http://bsscommerce.com/Bss-Commerce-License.txt
 */

namespace Bss\InventoryReport\Observer;

use Bss\InventoryReport\Helper\InventoryData as InventoryDataHelper;
use Bss\InventoryReport\Model\MultiInventorySource\MsiObjectManagement;
use Bss\InventoryReport\Model\MultiInventorySource\StockAndSourceProcessor;
use Bss\InventoryReport\Model\ReportFactory;
use Bss\InventoryReport\Model\ResourceModel\Report as InventoryResourceReport;
use Magento\Catalog\Api\ProductRepositoryInterface;
use Magento\CatalogInventory\Api\Data\StockItemInterface;
use Magento\CatalogInventory\Api\StockRegistryInterface;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\App\ProductMetadataInterface;
use Magento\Framework\Event\Observer;
use Magento\Framework\Event\ObserverInterface;
use Magento\Framework\Exception\InputException;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Framework\Stdlib\DateTime\DateTime;
use Magento\Framework\Stdlib\DateTime\TimezoneInterface;
use Magento\Sales\Api\OrderRepositoryInterface;
use Magento\Sales\Model\Order;
use Magento\Sales\Model\Order\Item as OrderItem;
use Magento\Store\Model\StoreManagerInterface;

class OrderCreateObserver implements ObserverInterface
{
    /**
     * @var DateTime
     */
    protected $date;

    /**
     * @var StockRegistryInterface
     */
    protected $stockRegistry;

    /**
     * @var InventoryResourceReport
     */
    protected $inventoryResourceReport;

    /**
     * @var OrderRepositoryInterface
     */
    protected $orderSale;

    /**
     * @var MsiObjectManagement
     */
    protected $msiObjectManagement;

    /**
     * @var ScopeConfigInterface
     */
    protected $scopeConfig;

    /**
     * @var InventoryDataHelper
     */
    protected $inventoryDataHelper;

    /**
     * @var ProductRepositoryInterface
     */
    protected $productRepository;

    /**
     * @var StoreManagerInterface
     */
    protected $storeManagerInterface;

    /**
     * @var StockAndSourceProcessor
     */
    protected $stockAndSourceProcessor;

    /**
     * @var TimezoneInterface
     */
    protected $timezone;

    /**
     * @var ReportFactory
     */
    protected $reportFactory;

    /**
     * @var ProductMetadataInterface
     */
    protected $productMetadata;

    /**
     * OrderCreateObserver constructor.
     * @param DateTime $date
     * @param StockRegistryInterface $stockRegistry
     * @param InventoryResourceReport $inventoryResourceReport
     * @param OrderRepositoryInterface $orderSale
     * @param MsiObjectManagement $msiObjectManagement
     * @param ScopeConfigInterface $scopeConfig
     * @param InventoryDataHelper $inventoryDataHelper
     * @param StoreManagerInterface $storeManagerInterface
     * @param ProductRepositoryInterface $productRepository
     * @param StockAndSourceProcessor $stockAndSourceProcessor
     * @param TimezoneInterface $timezone
     * @param ReportFactory $reportFactory
     * @param ProductMetadataInterface $productMetadata
     */
    public function __construct(
        DateTime $date,
        StockRegistryInterface $stockRegistry,
        InventoryResourceReport $inventoryResourceReport,
        OrderRepositoryInterface $orderSale,
        MsiObjectManagement $msiObjectManagement,
        ScopeConfigInterface $scopeConfig,
        InventoryDataHelper $inventoryDataHelper,
        StoreManagerInterface $storeManagerInterface,
        ProductRepositoryInterface $productRepository,
        StockAndSourceProcessor $stockAndSourceProcessor,
        TimezoneInterface $timezone,
        ReportFactory $reportFactory,
        ProductMetadataInterface $productMetadata
    ) {
        $this->date = $date;
        $this->stockRegistry = $stockRegistry;
        $this->inventoryResourceReport = $inventoryResourceReport;
        $this->orderSale = $orderSale;
        $this->msiObjectManagement = $msiObjectManagement;
        $this->scopeConfig = $scopeConfig;
        $this->inventoryDataHelper = $inventoryDataHelper;
        $this->storeManagerInterface = $storeManagerInterface;
        $this->productRepository = $productRepository;
        $this->stockAndSourceProcessor = $stockAndSourceProcessor;
        $this->timezone = $timezone;
        $this->reportFactory = $reportFactory;
        $this->productMetadata = $productMetadata;
    }

    /**
     * @param Observer $observer
     * @throws InputException
     * @throws LocalizedException
     * @throws NoSuchEntityException
     */
    public function execute(Observer $observer)
    {
        /** @var Order $order */
        $order = $observer->getEvent()->getOrder();
        // Only process this if MSI isn't enabled
        if (!$this->msiObjectManagement->isMsiEnabled()) {
            $this->processWithoutMsi($order);
        } else {
            $this->processWithMsi($order);
        }
    }

    /**
     * @param Order $order
     * @throws NoSuchEntityException
     */
    protected function processWithoutMsi($order)
    {
        $orderNumber = $order->getIncrementId();
        $items = $order->getAllItems();

        $customerId = $order->getCustomerId();
        $customerName = $this->getCustomerName($order, $customerId);
        $currentTime = $this->date->gmtDate();

        foreach ($items as $item) {
            if ($this->isAllowProductType($item)) {
                $productId = $item->getProductId();
                $productName = $item->getName();
                $productSku = $item->getSku();
                $qtyChange = $item->getQtyOrdered();

                $stockItem = $this->stockRegistry->getStockItemBySku($productSku);
                $stockQtyAfter = $stockItem->getQty();
                $stockStatus = $this->getStockStatus($stockItem, $stockQtyAfter, $productSku);

                $note = $this->getNote($orderNumber);
                $qtyChange = 0 - $qtyChange;
                $this->inventoryResourceReport->saveReport(
                    $currentTime,
                    $productId,
                    $productName,
                    $productSku,
                    $stockQtyAfter,
                    $qtyChange,
                    $stockStatus,
                    $customerName,
                    $note,
                    $customerId,
                    $orderNumber
                );
            }
        }
    }

    /**
     * @param Order $order
     * @param string|int $customerId
     * @return string
     */
    protected function getCustomerName($order, $customerId)
    {
        return $customerId ? $order->getBillingAddress()->getName() : __('Guest');
    }

    /**
     * @param OrderItem $orderItem
     * @return bool
     */
    protected function isAllowProductType($orderItem)
    {
        $productType = $orderItem->getProductType();
        return $productType == 'simple' || $productType == 'virtual' || $productType == 'grouped';
    }

    /**
     * @param StockItemInterface $stockItem
     * @param string|int $stockQtyAfter
     * @param string $productSku
     * @return int|string
     */
    protected function getStockStatus($stockItem, $stockQtyAfter, $productSku)
    {
        $backOrderConfigProduct = $stockItem->getBackorders();
        $useMinQtyConfig = $stockItem->getUseConfigMinQty();
        $minQtyConfigProduct = $stockItem->getMinQty();

        $backOrderValue = $this->inventoryDataHelper->getBackOrderValue();
        $threshold = $this->inventoryDataHelper->getThresholdValue();

        $stockStatus = $this->stockRegistry->getProductStockStatusBySku($productSku);

        if ($backOrderValue != 0 || $backOrderConfigProduct != 0) {
            $stockStatus = 1;
        } elseif ($stockStatus == 1 && $stockQtyAfter > 0) {
            if ($useMinQtyConfig && $stockQtyAfter > $threshold) {
                $stockStatus = 1;
            } elseif ($stockQtyAfter > $minQtyConfigProduct) {
                $stockStatus = 1;
            } else {
                $stockStatus = 0;
            }
        } else {
            $stockStatus = 0;
        }
        return $stockStatus;
    }

    /**
     * @param string $orderNumber
     * @return string
     */
    protected function getNote($orderNumber)
    {
        return __('Order Creation No: %1', $orderNumber);
    }

    /**
     * @param Order $order
     * @throws InputException
     * @throws LocalizedException
     * @throws NoSuchEntityException
     */
    protected function processWithMsi($order)
    {
        $data = [];
        $data['time'] = $this->timezone->date();
        $data['user'] = $this->getCustomerName($order, $order->getCustomerId());
        $data['note'] = $this->getNote($order->getIncrementId());
        $stockId = $this->getStockId($order);
        $stockLinkCollection = $this->stockAndSourceProcessor->getSourceProductByStockId($stockId);
        $checkSkip = [];
        foreach ($stockLinkCollection as $stockLink) {
            $sourceCode = $stockLink->getSourceCode();
            $data['source_code'] = $sourceCode;
            $this->processRecordOrder(
                $order,
                $data,
                $sourceCode,
                $stockId,
                $checkSkip
            );
        }
    }

    /**
     * @param Order $order
     * @return int|string
     * @throws NoSuchEntityException
     */
    protected function getStockId($order)
    {
        $store = $this->storeManagerInterface->getStore($order->getStoreId());
        $websiteCode = $store->getWebsite()->getCode();
        $stock = $this->stockAndSourceProcessor->getStockByWebsiteCode($websiteCode);
        return $stock->getStockId();
    }

    /**
     * @param mixed $order
     * @param array $data
     * @param string $sourceCode
     * @param int $stockId
     * @param array $checkSkip
     * @throws InputException
     * @throws LocalizedException
     * @throws NoSuchEntityException
     */
    protected function processRecordOrder(
        $order,
        $data,
        $sourceCode,
        $stockId,
        &$checkSkip
    )
    {
        $orderItems = $order->getAllVisibleItems();
        foreach ($orderItems as $item) {
            if ($this->isAllowProductType($item)) {
                $productSku = $item->getSku();
                if (isset($checkSkip[$productSku])) {
                    continue;
                }
                $sourceItem = $this->stockAndSourceProcessor->getSourceItem($productSku, $sourceCode);
                if ($sourceItem && $sourceItem->getQuantity() > 0) {
                    $checkSkip[$productSku] = true;
                    $product = $this->productRepository->get($productSku);
                    $salableQty = $this->stockAndSourceProcessor->getSaleableQtyByStockId($product->getSku(), $stockId);
                    $data['product_id'] = $product->getId();
                    $data['product_name'] = $product->getName();
                    $data['product_sku'] = $product->getSku();
                    $data['qty_after'] = $sourceItem->getQuantity();
                    $data['qty_change'] = null;
                    $data['saleable_qty'] = $this->getSalableQty($salableQty, $item);
                    $data['saleable_qty_change'] = -$item->getQtyOrdered();
                    $data['stock_status'] = $sourceItem->getStatus();
                    $data['order_increment_id'] = $order->getIncrementId();
                    $data['stock_id'] = $stockId;
                    $data['customer_id'] = $order->getCustomerId();
                    $this->reportFactory->create()->setData($data)->save();
                }
            }
        }
    }

    /**
     * @param float $salableQty
     * @param Order\Item $item
     * @return int
     */
    protected function getSalableQty($salableQty, $item)
    {
        if ($this->checkVersion24()) {
            return $salableQty ? $salableQty : 0;
        }
        $qty = $salableQty - $item->getQtyOrdered();
        return $qty > 0 ? $qty : 0;
    }

    /**
     * @return bool
     */
    private function checkVersion24()
    {
        $magentoVersion = $this->productMetadata->getVersion();
        return version_compare($magentoVersion, '2.4.0', '>=');
    }
}
