<?php
/**
 * BSS Commerce Co.
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the EULA
 * that is bundled with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://bsscommerce.com/Bss-Commerce-License.txt
 *
 * @category   BSS
 * @package    Bss_InventoryReport
 * @author     Extension Team
 * @copyright  Copyright (c) 2017-2018 BSS Commerce Co. ( http://bsscommerce.com )
 * @license    http://bsscommerce.com/Bss-Commerce-License.txt
 */

namespace Bss\InventoryReport\Observer;

use Bss\InventoryReport\Helper\InventoryData as InventoryDataHelper;
use Bss\InventoryReport\Model\MultiInventorySource\MsiObjectManagement;
use Bss\InventoryReport\Model\MultiInventorySource\StockAndSourceProcessor;
use Bss\InventoryReport\Model\ReportFactory;
use Magento\Catalog\Api\Data\ProductInterface;
use Magento\Catalog\Api\ProductRepositoryInterface;
use Magento\Catalog\Model\Product\Type as ProductType;
use Magento\Downloadable\Model\Product\Type;
use Magento\Framework\Event\Observer;
use Magento\Framework\Event\ObserverInterface;
use Magento\Framework\Exception\InputException;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Framework\Stdlib\DateTime\TimezoneInterface;
use Magento\Sales\Model\Order\Invoice;
use Magento\Sales\Model\Order\Invoice\Item;
use Magento\Store\Api\WebsiteRepositoryInterface;

class OrderInvoiceObserver implements ObserverInterface
{
    /**
     * @var WebsiteRepositoryInterface
     */
    protected $websiteRepository;

    /**
     * @var MsiObjectManagement
     */
    protected $msiObjectManagement;

    /**
     * @var StockAndSourceProcessor
     */
    protected $stockAndSourceProcessor;

    /**
     * @var ReportFactory
     */
    protected $reportFactory;

    /**
     * @var TimezoneInterface
     */
    protected $timezone;

    /**
     * @var InventoryDataHelper
     */
    protected $inventoryDataHelper;

    /**
     * @var ProductRepositoryInterface
     */
    protected $productRepository;

    /**
     * OrderInvoiceObserver constructor.
     * @param WebsiteRepositoryInterface $websiteRepository
     * @param MsiObjectManagement $msiObjectManagement
     * @param StockAndSourceProcessor $stockAndSourceProcessor
     * @param ReportFactory $reportFactory
     * @param TimezoneInterface $timezone
     * @param InventoryDataHelper $inventoryDataHelper
     * @param ProductRepositoryInterface $productRepository
     */
    public function __construct(
        WebsiteRepositoryInterface $websiteRepository,
        MsiObjectManagement $msiObjectManagement,
        StockAndSourceProcessor $stockAndSourceProcessor,
        ReportFactory $reportFactory,
        TimezoneInterface $timezone,
        InventoryDataHelper $inventoryDataHelper,
        ProductRepositoryInterface $productRepository
    ) {
        $this->websiteRepository = $websiteRepository;
        $this->msiObjectManagement = $msiObjectManagement;
        $this->stockAndSourceProcessor = $stockAndSourceProcessor;
        $this->reportFactory = $reportFactory;
        $this->timezone = $timezone;
        $this->inventoryDataHelper = $inventoryDataHelper;
        $this->productRepository = $productRepository;
    }

    /**
     * Execute product qty when invoice order (virtual, download product type)
     *
     * @param Observer $observer
     * @return void
     * @throws InputException
     * @throws LocalizedException
     * @throws NoSuchEntityException
     */
    public function execute(Observer $observer)
    {
        // Only process this if MSI is enabled
        if ($this->msiObjectManagement->isMsiEnabled()) {
            /** @var Invoice $invoice */
            $invoice = $observer->getEvent()->getInvoice();
            $invoiceItems = $invoice->getAllItems();

            /** @var Item $invoiceItem */
            foreach ($invoiceItems as $invoiceItem) {
                $product = $this->productRepository->get($invoiceItem->getSku());
                if ($this->isAllowProductType($product)) {
                    $websiteCode = $this->getWebsiteCode($invoice);

                    $stockId = $this->stockAndSourceProcessor->getStockByWebsiteCode($websiteCode)->getStockId();
                    $sourceCode = $this->stockAndSourceProcessor->getSourceCodeByStockId($stockId);
                    $sourceItem = $this->stockAndSourceProcessor->getSourceItem($product->getSku(), $sourceCode);
                    $salableQty = $this->stockAndSourceProcessor->getSaleableQtyByStockId($product->getSku(), $stockId);

                    $data['source_code'] = $sourceCode;
                    $data['time'] = $this->timezone->date();
                    $data['note'] = $this->getNote($invoice->getIncrementId());
                    $data['user'] = $this->inventoryDataHelper->getAdminName();
                    $data['stock_id'] = $stockId;
                    $data['product_id'] = $product->getId();
                    $data['product_name'] = $product->getName();
                    $data['product_sku'] = $product->getSku();
                    $data['qty_change'] = 0 - $invoiceItem->getQty();
                    $data['qty_after'] = $sourceItem->getQuantity();
                    $data['stock_status'] = $sourceItem->getStatus();
                    $data['saleable_qty'] = $salableQty;
                    $data['saleable_qty_change'] = null;
                    $data['invoice_entity_id'] = $invoice->getId();
                    $this->reportFactory->create()->setData($data)->save();
                }
            }
        }
    }

    /**
     * @param ProductInterface $product
     * @return bool
     */
    protected function isAllowProductType($product)
    {
        return $product->getTypeId() == Type::TYPE_DOWNLOADABLE ||
            $product->getTypeId() == ProductType::TYPE_VIRTUAL;
    }

    /**
     * @param Invoice $invoice
     * @return string
     * @throws NoSuchEntityException
     */
    protected function getWebsiteCode($invoice)
    {
        $websiteId = $invoice->getStore()->getWebsiteId();
        return $this->websiteRepository->getById($websiteId)->getCode();
    }

    /**
     * @param string|int $invoiceIncrementId
     * @return string
     */
    protected function getNote($invoiceIncrementId)
    {
        return __('Order Invoiced: %1', $invoiceIncrementId);
    }
}
