<?php
/**
 * BSS Commerce Co.
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the EULA
 * that is bundled with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://bsscommerce.com/Bss-Commerce-License.txt
 *
 * @category   BSS
 * @package    Bss_InventoryReport
 * @author     Extension Team
 * @copyright  Copyright (c) 2017-2018 BSS Commerce Co. ( http://bsscommerce.com )
 * @license    http://bsscommerce.com/Bss-Commerce-License.txt
 */

namespace Bss\InventoryReport\Observer;

use Bss\InventoryReport\Helper\InventoryData as InventoryDataHelper;
use Bss\InventoryReport\Model\MultiInventorySource\MsiObjectManagement;
use Bss\InventoryReport\Model\ResourceModel\Report as InventoryResourceReport;
use Magento\Catalog\Api\Data\ProductInterface;
use Magento\Catalog\Model\ProductRepository;
use Magento\CatalogInventory\Api\Data\StockItemInterface;
use Magento\CatalogInventory\Api\StockRegistryInterface;
use Magento\Framework\Event\Observer;
use Magento\Framework\Event\ObserverInterface;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Framework\Stdlib\DateTime\DateTime;
use Magento\Sales\Model\Order\Creditmemo;

class OrderRefundObserver implements ObserverInterface
{
    /**
     * @var DateTime
     */
    protected $date;

    /**
     * @var StockRegistryInterface
     */
    protected $stockRegistry;

    /**
     * @var InventoryResourceReport
     */
    protected $inventoryResourceReport;

    /**
     * @var ProductRepository
     */
    protected $productRepository;

    /**
     * @var MsiObjectManagement
     */
    protected $msiObjectManagement;

    /**
     * @var InventoryDataHelper
     */
    protected $inventoryDataHelper;

    /**
     * OrderRefundObserver constructor.
     * @param DateTime $date
     * @param StockRegistryInterface $stockRegistry
     * @param InventoryResourceReport $inventoryResourceReport
     * @param ProductRepository $productRepository
     * @param MsiObjectManagement $msiObjectManagement
     * @param InventoryDataHelper $inventoryDataHelper
     */
    public function __construct(
        DateTime $date,
        StockRegistryInterface $stockRegistry,
        InventoryResourceReport $inventoryResourceReport,
        ProductRepository $productRepository,
        MsiObjectManagement $msiObjectManagement,
        InventoryDataHelper $inventoryDataHelper
    ) {
        $this->date = $date;
        $this->stockRegistry = $stockRegistry;
        $this->inventoryResourceReport = $inventoryResourceReport;
        $this->productRepository = $productRepository;
        $this->msiObjectManagement = $msiObjectManagement;
        $this->inventoryDataHelper = $inventoryDataHelper;
    }

    /**
     * @param Observer $observer
     * @throws NoSuchEntityException
     */
    public function execute(Observer $observer)
    {
        // Only process this if MSI isn't enabled
        if (!$this->msiObjectManagement->isMsiEnabled()) {
            /** @var Creditmemo $creditMemo */
            $creditMemo = $observer->getEvent()->getCreditmemo();
            $creditId = $creditMemo->getData('entity_id');
            $items = $creditMemo->getAllItems();

            $currentTime = $this->date->gmtDate();
            $user = $this->inventoryDataHelper->getAdminName();

            foreach ($items as $item) {
                $backToStock = $item->getData('back_to_stock');
                if ($backToStock) {
                    $product = $this->productRepository->getById($item->getProductId());
                    if ($this->isAllowProductType($product)) {
                        $productName = $product->getName();
                        $productId = $product->getId();
                        $productSku = $product->getSku();
                        $productType = $product->getTypeId();

                        $stockItem = $this->stockRegistry->getStockItemBySku($productSku);
                        $stockQtyAfter = $stockItem->getQty();
                        $stockStatus = $this->getStatus(
                            $stockItem,
                            $stockQtyAfter,
                            $productSku
                        );

                        $qtyChange = $item->getQty();
                        $note = $this->getNote($creditMemo->getIncrementId());

                        $this->inventoryResourceReport->saveReport(
                            $currentTime,
                            $productId,
                            $productName,
                            $productSku,
                            $stockQtyAfter,
                            $qtyChange,
                            $stockStatus,
                            $user,
                            $note,
                            null,
                            $creditId
                        );
                    }
                }
            }
        }
    }

    /**
     * @param ProductInterface $product
     * @return bool
     */
    protected function isAllowProductType($product)
    {
        $productType = $product->getTypeId();
        return $productType == 'simple' || $productType == 'virtual';
    }

    /**
     * @param StockItemInterface $stockItem
     * @param string|int $stockQtyAfter
     * @param string $productSku
     * @return int|string
     */
    protected function getStatus($stockItem, $stockQtyAfter, $productSku)
    {
        $useMinQtyConfig = $stockItem->getUseConfigMinQty();
        $minQtyConfigProduct = $stockItem->getMinQty();

        $threshold = $this->inventoryDataHelper->getThresholdValue();
        $stockStatus = $this->stockRegistry->getProductStockStatusBySku($productSku);

        if ($stockStatus == 1 || $stockQtyAfter > 0) {
            if ($useMinQtyConfig && $stockQtyAfter > $threshold) {
                $stockStatus = 1;
            } elseif ($stockQtyAfter > $minQtyConfigProduct) {
                $stockStatus = 1;
            } else {
                $stockStatus = 0;
            }
        } else {
            $stockStatus = 0;
        }
        return $stockStatus;
    }

    /**
     * @param string|int $creditIncrement
     * @return string
     */
    protected function getNote($creditIncrement)
    {
        return __('Credit Memo Refund No: %1', $creditIncrement);
    }
}
