<?php
/**
 * BSS Commerce Co.
 *
 * NOTICE OF LICENSE
 *
 * @category   BSS
 * @package    Bss_InventoryReport
 * @author     Extension Team
 * @copyright  Copyright (c) 2018-2019 BSS Commerce Co. ( http://bsscommerce.com )
 * @license    http://bsscommerce.com/Bss-Commerce-License.txt
 */

namespace Bss\InventoryReport\Observer;

use Bss\InventoryReport\Helper\InventoryData as InventoryDataHelper;
use Bss\InventoryReport\Model\MultiInventorySource\MsiObjectManagement;
use Bss\InventoryReport\Model\MultiInventorySource\StockAndSourceProcessor;
use Bss\InventoryReport\Model\ReportFactory;
use Bss\InventoryReport\Model\ResourceModel\Report as InventoryResourceReport;
use Exception;
use Magento\Catalog\Api\Data\ProductInterface;
use Magento\Catalog\Model\Product;
use Magento\Catalog\Model\Product\Type;
use Magento\CatalogInventory\Api\StockRegistryInterface;
use Magento\CatalogInventory\Model\Stock\Status;
use Magento\Framework\App\ProductMetadataInterface;
use Magento\Framework\App\RequestInterface;
use Magento\Framework\Event\Observer;
use Magento\Framework\Event\ObserverInterface;
use Magento\Framework\Registry;
use Magento\Framework\Stdlib\DateTime\DateTime;
use Magento\Framework\Stdlib\DateTime\TimezoneInterface;

class SaveAfterObserver implements ObserverInterface
{
    /**
     * @var RequestInterface
     */
    protected $request;
    /**
     * @var Registry
     */
    protected $registry;

    /**
     * @var MsiObjectManagement
     */
    protected $msiObjectManagement;

    /**
     * @var StockAndSourceProcessor
     */
    protected $stockAndSourceProcessor;

    /**
     * @var TimezoneInterface
     */
    protected $timezone;

    /**
     * @var ReportFactory
     */
    protected $reportFactory;

    /**
     * @var DateTime
     */
    protected $dateTime;

    /**
     * @var StockRegistryInterface
     */
    protected $stockRegistry;

    /**
     * @var ProductMetadataInterface
     */
    protected $productMetadata;

    /**
     * @var InventoryResourceReport
     */
    protected $inventoryResourceReport;

    /**
     * @var InventoryDataHelper
     */
    protected $inventoryDataHelper;

    /**
     * SaveAfterObserver constructor.
     * @param RequestInterface $request
     * @param Registry $registry
     * @param MsiObjectManagement $msiObjectManagement
     * @param StockAndSourceProcessor $stockAndSourceProcessor
     * @param TimezoneInterface $timezone
     * @param ReportFactory $reportFactory
     * @param DateTime $dateTime
     * @param StockRegistryInterface $stockRegistry
     * @param ProductMetadataInterface $productMetadata
     * @param InventoryResourceReport $inventoryResourceReport
     * @param InventoryDataHelper $inventoryDataHelper
     */
    public function __construct(
        RequestInterface $request,
        Registry $registry,
        MsiObjectManagement $msiObjectManagement,
        StockAndSourceProcessor $stockAndSourceProcessor,
        TimezoneInterface $timezone,
        ReportFactory $reportFactory,
        Datetime $dateTime,
        StockRegistryInterface $stockRegistry,
        ProductMetadataInterface $productMetadata,
        InventoryResourceReport $inventoryResourceReport,
        InventoryDataHelper $inventoryDataHelper
    ) {
        $this->request = $request;
        $this->registry = $registry;
        $this->msiObjectManagement = $msiObjectManagement;
        $this->stockAndSourceProcessor = $stockAndSourceProcessor;
        $this->timezone = $timezone;
        $this->reportFactory = $reportFactory;
        $this->dateTime = $dateTime;
        $this->stockRegistry = $stockRegistry;
        $this->productMetadata = $productMetadata;
        $this->inventoryResourceReport = $inventoryResourceReport;
        $this->inventoryDataHelper = $inventoryDataHelper;
    }

    /**
     * Process Save Report Inventory After Save Product
     *
     * @param Observer $observer
     * @throws Exception
     */
    public function execute(Observer $observer)
    {
        $product = $observer->getEvent()->getProduct();
        if ($this->msiObjectManagement->isMsiEnabled()) {
            $this->saveWithMsi($product);
        }
        $this->registry->unregister('admin_save_product');
        $this->registry->register('admin_save_product', '1');
    }

    /**
     * @param Product|ProductInterface $product
     * @throws Exception
     */
    protected function saveWithMsi($product)
    {
        $data = [];
        /** @var Product $product */
        $posts = $this->request->getPost();
        $minQty = $this->getMinQty($posts);

        $data['time'] = $this->timezone->date();
        $data['user'] = $this->inventoryDataHelper->getAdminName();
        $data['note'] = __('Admin Manual Change');
        $data['product_id'] = $product->getId();
        $data['product_name'] = $product->getName();
        $data['product_sku'] = $product->getSku();

        $sourceItems = $this->stockAndSourceProcessor->getListSource($product->getSku());
        $unassignSources = $this->processGetListSource($sourceItems);
        $countUnassgin = count($unassignSources);
        $countSourcePost = 0;
        $assignSource = $this->registry->registry('bss_source_qty_before');

        if (isset($posts['sources']) &&
            isset($posts['sources']['assigned_sources']) &&
            $assignSource) {
            $assignedSources = $posts['sources']['assigned_sources'];
            $countSourcePost = count($assignedSources);
            foreach ($assignedSources as $source) {
                $sourceCode = isset($source['source_code']) ? $source['source_code'] : 'default';
                $unassignSources = $this->removeSource($unassignSources, $sourceCode);
                if ($this->checkAllowProductType($product->getTypeId())) {
                    $this->processDataForMsi(
                        $data,
                        $source,
                        $sourceCode,
                        $assignSource,
                        $minQty
                    );
                }
            }
        } else {
            if ($this->checkAllowProductType($product->getTypeId())) {
                $this->processMsiSingleDefault($data, $posts, $product);
            }
            $countUnassgin = 0;
        }
        /** save report for unassginSource **/
        $this->processUnassignStorageData(
            $unassignSources,
            $data,
            $countSourcePost,
            $countUnassgin
        );
    }

    /**
     * @param array $postValue
     * @return float|int
     */
    protected function getMinQty($postValue)
    {
        $afterMinQty = 0;
        if (isset($postValue['product']['stock_data'])) {
            $stockData = $postValue['product']['stock_data'];
            if (isset($stockData['min_qty'])) {
                $afterMinQty = $stockData['min_qty'];
            }
        }
        return $afterMinQty;
    }

    /**
     * Get List source
     *
     * @param object $sourceItems
     * @return array
     */
    protected function processGetListSource($sourceItems)
    {
        $unassignSources = [];
        foreach ($sourceItems as $sourceItem) {
            $unassignSources[$sourceItem->getSourceCode()] = $sourceItem->getQuantity();
        }
        return $unassignSources;
    }

    /**
     * Remove source from unassign source
     *
     * @param array $unassignSources
     * @param string $sourceCode
     * @return array
     */
    protected function removeSource($unassignSources, $sourceCode)
    {
        if (isset($unassignSources[$sourceCode])) {
            unset($unassignSources[$sourceCode]);
        }
        return $unassignSources;
    }

    /**
     * @param string $type
     * @return bool
     */
    protected function checkAllowProductType($type)
    {
        $notAllowType = [
            Type::TYPE_BUNDLE,
            'configurable',
            'grouped'
        ];
        return !in_array($type, $notAllowType);
    }

    /**
     * Process Report for Msi
     *
     * @param array $data
     * @param array $source
     * @param string $sourceCode
     * @param array $assignSource
     * @param float $minQty
     * @throws Exception
     */
    protected function processDataForMsi(
        $data,
        $source,
        $sourceCode,
        $assignSource,
        $minQty
    ) {
        $qtyAfter = $source['quantity'];
        $stockStatus = $source['status'];
        $productSku = $data['product_sku'];
        $data['source_code'] = $sourceCode;
        $stockStatusBefore = $stockStatus;

        $qtyChange = $this->getQtyChange($assignSource, $sourceCode, $qtyAfter);
        $sourceStocks = $this->stockAndSourceProcessor->getStockBySourceCode($sourceCode);
        if ($this->stockAndSourceProcessor->getSourceItem($productSku, $sourceCode)) {
            $stockStatusBefore = $this->stockAndSourceProcessor->getSourceItem(
                $productSku,
                $sourceCode
            )->getStatus();
        }

        foreach ($sourceStocks as $sourceStock) {
            $stockId = $sourceStock->getStockId();
            $qtyReservation = $this->stockAndSourceProcessor->getReservationQty($productSku, $stockId);
            $minQtyChange = $minQty - (float)$this->registry->registry('bss_product_min_qty_before');
            $salableBefore = $this->stockAndSourceProcessor->getSaleableQtyByStockId($productSku, $stockId);
            $salableQtyChange = $this->getSalableQtyChange(
                $stockStatusBefore,
                $stockStatus,
                $salableBefore,
                $qtyReservation,
                $qtyChange,
                $minQtyChange,
                $qtyAfter
            );
            $salableQty = $salableBefore + $salableQtyChange;
            if ((int)$qtyChange !== 0 || $stockStatus != $stockStatusBefore) {
                $data['qty_change'] = $qtyChange;
                $data['stock_status'] = $stockStatus;
                $data['qty_after'] = $qtyAfter;
                $data['saleable_qty'] = $salableQty;
                $data['saleable_qty_change'] = $salableQtyChange;
                $data['stock_id'] = $stockId;
                $this->reportFactory->create()->setData($data)->save();
            }
        }
    }

    /**
     * @param array $assignSource
     * @param string $sourceCode
     * @param int|float|string $qtyAfter
     * @return int|float|string
     */
    protected function getQtyChange($assignSource, $sourceCode, $qtyAfter)
    {
        $qtyBefore = $qtyAfter;
        if (isset($assignSource[$sourceCode]['qty'])) {
            $qtyBefore = $assignSource[$sourceCode]['qty'];
        }
        $qtyChange = $qtyAfter - $qtyBefore;
        return $qtyChange;
    }

    /**
     * @param int|bool|string $stockStatusBefore
     * @param int|bool|string $stockStatus
     * @param int|float|string $salableBefore
     * @param int|float|string $qtyReservation
     * @param int|float|string $qtyChange
     * @param int|float|string $minQtyChange
     * @param int|float|string $qtyAfter
     * @return float|int
     */
    protected function getSalableQtyChange(
        $stockStatusBefore,
        $stockStatus,
        $salableBefore,
        $qtyReservation,
        $qtyChange,
        $minQtyChange,
        $qtyAfter
    ) {
        if ($stockStatus == Status::STATUS_OUT_OF_STOCK) {
            $salableQtyChange = 0 - $salableBefore;
        } elseif ($qtyChange && $stockStatusBefore == Status::STATUS_IN_STOCK) {
            $salableQtyChange = ($qtyChange + $qtyReservation) - $minQtyChange;
        } else {
            $salableQtyChange = (float)$qtyAfter;
        }
        return $salableQtyChange;
    }

    /**
     * Process report for msi single default store
     *
     * @param array $data
     * @param array $posts
     * @param Product $product
     * @throws Exception
     */
    protected function processMsiSingleDefault($data, $posts, $product)
    {
        //Process Single Default source MSI
        $data['source_code'] = 'default';
        $qtyAfter = $qtyBefore = 0;
        $stockStatus = Status::STATUS_OUT_OF_STOCK;
        $statusBefore = $stockStatus;
        if (isset($posts['product']['stock_data']['is_in_stock'])) {
            $statusBefore = $posts['product']['stock_data']['is_in_stock'];
        }
        if (isset($posts['product']['quantity_and_stock_status'])) {
            if (!isset($posts['product']['quantity_and_stock_status']['qty'])) {
                $data['stock_status'] = $posts['product']['quantity_and_stock_status']['is_in_stock'];
                $qtyAfter = (float)$posts['product']['stock_data']['qty'];
            } else {
                $qtyAfter = (float)$posts['product']['quantity_and_stock_status']['qty'];
            }
        }

        $qtyAfter = $this->getQtyAfter($posts, $qtyAfter);
        $qtyBefore = $this->getQtyBefore($posts, $qtyBefore);
        $qtyChange = $qtyAfter - $qtyBefore;

        if (!$qtyChange || $qtyChange == 0) {
            $qtySingleSourceBefore = $this->registry->registry('bss_qty_single_default_stock_before');
            if ($qtySingleSourceBefore) {
                $qtyChange = $qtyAfter - (float)$qtySingleSourceBefore;
            }
        }

        $stockItem = $product->getExtensionAttributes()->getStockItem();
        if ($stockItem) {
            $stockStatus = (int)$stockItem->getIsInStock();
        }

        $salableQtyBefore = 0;
        if (isset($this->registry->registry('bss_salable_qty_before')[0]['qty'])) {
            $salableQtyBefore = $this->registry->registry('bss_salable_qty_before')[0]['qty'];
        }
        $salableQtyChange = $this->getSalableQtyChangeDefault(
            $salableQtyBefore,
            $statusBefore,
            $stockStatus,
            $qtyAfter,
            $qtyChange
        );
        $salableQty = $salableQtyBefore + $salableQtyChange;
        $data['qty_after'] = $qtyAfter;
        $data['qty_change'] = $qtyChange;
        $data['saleable_qty'] = $salableQty;
        $data['saleable_qty_change'] = $salableQtyChange;
        $data['stock_status'] = $stockStatus;
        if ((int)$qtyChange !== 0 || $stockStatus != $statusBefore) {
            $this->reportFactory->create()->setData($data)->save();
        }
    }

    /**
     * @param array $posts
     * @param int|float|string $qtyAfter
     * @return float
     */
    protected function getQtyAfter($posts, $qtyAfter)
    {
        if (isset($posts['product']['quantity_and_stock_status'])) {
            if (!isset($posts['product']['quantity_and_stock_status']['qty'])) {
                $qtyAfter = (float)$posts['product']['stock_data']['qty'];
            } else {
                $qtyAfter = (float)$posts['product']['quantity_and_stock_status']['qty'];
            }
        }
        return $qtyAfter;
    }

    /**
     * @param array $posts
     * @param int|float|string $qtyBefore
     * @return float
     */
    protected function getQtyBefore($posts, $qtyBefore)
    {
        if (isset($posts['product']['stock_data']['qty'])) {
            $qtyBefore = (float)$posts['product']['stock_data']['qty'];
        }
        return $qtyBefore;
    }

    /**
     * @param int|float|string $salableQtyBefore
     * @param int|float|string $statusBefore
     * @param string|bool|int $stockStatus
     * @param int|float|string $qtyAfter
     * @param int|float|string $qtyChange
     * @return float|int
     */
    protected function getSalableQtyChangeDefault(
        $salableQtyBefore,
        $statusBefore,
        $stockStatus,
        $qtyAfter,
        $qtyChange
    ) {
        if ($stockStatus == Status::STATUS_OUT_OF_STOCK) {
            $salableQtyChange = 0 - $salableQtyBefore;
        } elseif ($qtyChange && $statusBefore == Status::STATUS_IN_STOCK) {
            $salableQtyChange = $qtyChange;
        } else {
            $salableQtyChange = (float)$qtyAfter;
        }
        return $salableQtyChange;
    }

    /**
     * Process Save Report to database
     *
     * @param array $unassignSources
     * @param array $data
     * @param int $countSourcePost
     * @param int $countUnassgin
     * @throws Exception
     */
    protected function processUnassignStorageData(
        $unassignSources,
        $data,
        $countSourcePost,
        $countUnassgin
    ) {
        if (!empty($unassignSources) && ($countSourcePost < $countUnassgin)) {
            foreach ($unassignSources as $sourceCode => $unassignSource) {
                if ($unassignSource > 0) {
                    $data['note'] = __('Admin Unassign Source');
                    $data['source_code'] = $sourceCode;
                    $data['qty_change'] = 0 - $unassignSource;
                    $data['stock_status'] = 0;
                    $data['qty_after'] = 0;
                    $this->reportFactory->create()->setData($data)->save();
                }
            }
        }
    }
}
