<?php
/**
 * BSS Commerce Co.
 *
 * NOTICE OF LICENSE
 *
 * @category   BSS
 * @package    Bss_InventoryReport
 * @author     Extension Team
 * @copyright  Copyright (c) 2018-2019 BSS Commerce Co. ( http://bsscommerce.com )
 * @license    http://bsscommerce.com/Bss-Commerce-License.txt
 */

namespace Bss\InventoryReport\Observer;

use Bss\InventoryReport\Helper\InventoryData as InventoryDataHelper;
use Bss\InventoryReport\Model\MultiInventorySource\MsiObjectManagement;
use Bss\InventoryReport\Model\MultiInventorySource\StockAndSourceProcessor;
use Bss\InventoryReport\Model\ResourceModel\Report as InventoryResourceReport;
use Exception;
use Magento\Catalog\Api\Data\ProductInterface;
use Magento\Catalog\Api\ProductRepositoryInterface;
use Magento\Catalog\Model\Product;
use Magento\CatalogInventory\Api\StockRegistryInterface;
use Magento\Framework\App\ProductMetadataInterface;
use Magento\Framework\App\RequestInterface;
use Magento\Framework\Event\Observer;
use Magento\Framework\Event\ObserverInterface;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Framework\Registry;
use Magento\Framework\Stdlib\DateTime\DateTime;

class SaveObserver implements ObserverInterface
{
    /**
     * @var Registry
     */
    protected $registry;

    /**
     * @var RequestInterface
     */
    protected $request;

    /**
     * @var StockAndSourceProcessor
     */
    protected $stockAndSourceProcessor;

    /**
     * @var MsiObjectManagement
     */
    protected $msiObjectManagement;

    /**
     * @var StockRegistryInterface
     */
    protected $stockRegistry;

    /**
     * @var ProductMetadataInterface
     */
    protected $productMetadata;

    /**
     * @var InventoryResourceReport
     */
    protected $inventoryResourceReport;

    /**
     * @var DateTime
     */
    protected $dateTime;

    /**
     * @var InventoryDataHelper
     */
    protected $inventoryDataHelper;

    /**
     * @var ProductRepositoryInterface
     */
    protected $productRepository;

    /**
     * SaveObserver constructor.
     * @param Registry $registry
     * @param RequestInterface $request
     * @param StockAndSourceProcessor $stockAndSourceProcessor
     * @param MsiObjectManagement $msiObjectManagement
     * @param StockRegistryInterface $stockRegistry
     * @param ProductMetadataInterface $productMetadata
     * @param InventoryResourceReport $inventoryResourceReport
     * @param DateTime $dateTime
     * @param InventoryDataHelper $inventoryDataHelper
     * @param ProductRepositoryInterface $productRepository
     */
    public function __construct(
        Registry $registry,
        RequestInterface $request,
        StockAndSourceProcessor $stockAndSourceProcessor,
        MsiObjectManagement $msiObjectManagement,
        StockRegistryInterface $stockRegistry,
        ProductMetadataInterface $productMetadata,
        InventoryResourceReport $inventoryResourceReport,
        DateTime $dateTime,
        InventoryDataHelper $inventoryDataHelper,
        ProductRepositoryInterface $productRepository
    ) {
        $this->request = $request;
        $this->registry = $registry;
        $this->stockAndSourceProcessor = $stockAndSourceProcessor;
        $this->msiObjectManagement = $msiObjectManagement;
        $this->stockRegistry = $stockRegistry;
        $this->productMetadata = $productMetadata;
        $this->inventoryResourceReport = $inventoryResourceReport;
        $this->dateTime = $dateTime;
        $this->inventoryDataHelper = $inventoryDataHelper;
        $this->productRepository = $productRepository;
    }

    /**
     * @param Observer $observer
     * @return $this|void
     */
    public function execute(Observer $observer)
    {
        $product = $observer->getEvent()->getProduct();
        if ($this->msiObjectManagement->isMsiEnabled()) {
            $this->saveWithMsi($product);
        } else {
            $this->saveWithoutMsi($product);
        }

        $this->registry->unregister('admin_save_product');
        $this->registry->register('admin_save_product', '1');
        return $this;
    }

    /**
     * @param Product|ProductInterface $product
     */
    protected function saveWithMsi($product)
    {
        $posts = $this->request->getPost();
        $productSku = $product->getSku();
        $assignSource = [];
        if (isset($posts['sources']) &&
            isset($posts['sources']['assigned_sources']) &&
            is_array($posts['sources']['assigned_sources'])) {
            $sources = $posts['sources']['assigned_sources'];
            foreach ($sources as $source) {
                $sourceCode = isset($source['source_code']) ? $source['source_code'] : 'default';
                $qtyBeforeChange = $this->getQtyBeforeChange($sourceCode, $productSku);
                $assignSource[$sourceCode]['qty'] = floatval($qtyBeforeChange);
                $sourceStocks = $this->stockAndSourceProcessor->getStockBySourceCode($sourceCode);

                foreach ($sourceStocks as $sourceStock) {
                    $stockId = $sourceStock->getStockId();
                    $qtySale = 0;
                    try {
                        if ($product->getId()) {
                            $qtySale = $this->stockAndSourceProcessor->getSaleableQtyByStockId($productSku, $stockId);
                        }
                    } catch (Exception $e) {
                        $qtySale = 0;
                    }
                    $assignSource[$sourceCode]['saleable'][$stockId] = $qtySale;
                }
            }
        } else {
            $this->registry->unregister('bss_qty_single_default_stock_before');
            $wsId = $posts['product']['stock_data']['website_id'] ?? null;
            $pId = $posts['product']['stock_data']['product_id'] ?? null;
            if ($wsId !== null && $pId) {
                $stockItem = $this->stockRegistry->getStockItem($pId);
                $qtyBeforeFirst = $stockItem->getQty();
                $this->registry->register('bss_qty_single_default_stock_before', $qtyBeforeFirst);
            }
        }
        $productId = $this->request->getParam('id');
        try {
            $productAfter = $this->productRepository->getById($productId);
            if ($productAfter->getSku() !== $productSku) {
                $productSku = $productAfter->getSku();
            }
        } catch (NoSuchEntityException $exception) {
            // do nothing
        }
        $salableQty = $this->stockAndSourceProcessor->getSaleableQty($productSku);
        $this->registry->unregister('bss_salable_qty_before');
        $this->registry->register('bss_salable_qty_before', $salableQty);
        $this->registry->unregister('bss_source_qty_before');
        $this->registry->register('bss_source_qty_before', $assignSource);
        $this->setMinQtyProduct($product);
    }

    /**
     * @param string $sourceCode
     * @param string $productSku
     * @return int
     */
    protected function getQtyBeforeChange($sourceCode, $productSku)
    {
        $sourceItem = $this->stockAndSourceProcessor->getSourceItem($productSku, $sourceCode);
        if (!$sourceItem) {
            return 0;
        }
        return $sourceItem->getQuantity();
    }

    /**
     * @param Product|ProductInterface $product
     */
    protected function setMinQtyProduct($product)
    {
        $notAllowType = [
            'bundle',
            'configurable',
            'grouped'
        ];
        $this->registry->unregister('bss_product_min_qty_before');
        if ($product->getId() && !in_array($product->getTypeId(), $notAllowType)) {
            $minQty = $product->getExtensionAttributes()->getStockItem()->getMinQty();
            $this->registry->register('bss_product_min_qty_before', $minQty);
        } else {
            $this->registry->register('bss_product_min_qty_before', 0);
        }
    }

    /**
     * @param Product|ProductInterface $product
     */
    protected function saveWithoutMsi($product)
    {
        $id = $this->request->getParam('id');
        if ($id && $this->isAllowProductType($product)) {
            $productId = $product->getId();
            $productName = $product->getName();
            $productSku = $product->getSku();

            $posts = $this->request->getPost();
            $stockData = $posts['product']['stock_data'] ?? [];
            $qtyStatusData = $posts['product']['quantity_and_stock_status'] ?? [];
            if (!empty($stockData) && !empty($qtyStatusData) && isset($qtyStatusData['qty'])) {
                $stockItem = $this->stockRegistry->getStockItem($productId);
                $stockQtyAfter = (float)$qtyStatusData['qty'];
                $stockQty = $stockItem->getQty();
                $qtyChange = (float)$stockQtyAfter - (float)$stockQty;
                $stockStatus = $this->checkStatus(
                    $qtyChange,
                    $stockQtyAfter,
                    $stockData
                );
                $admin = $this->inventoryDataHelper->getAdminName();
                $currentTime = $this->dateTime->gmtDate();
                $note = __('Admin Manual Change');
                if ($stockQtyAfter !== 0 && $qtyChange !== (float)0) {
                    $this->inventoryResourceReport->saveReport(
                        $currentTime,
                        $productId,
                        $productName,
                        $productSku,
                        $stockQtyAfter,
                        $qtyChange,
                        $stockStatus,
                        $admin,
                        $note,
                        null,
                        null
                    );
                }
            }
        }
    }

    /**
     * @param Product|ProductInterface $product
     * @return bool
     */
    protected function isAllowProductType($product)
    {
        $productType = $product->getTypeId();
        return $productType == 'simple' || $productType == 'virtual';
    }

    /**
     * @param int|float $qtyChange
     * @param int|float $qtyAfter
     * @param array $stockData
     * @return int
     */
    protected function checkStatus(
        $qtyChange,
        $qtyAfter,
        $stockData
    ) {
        $backOrderValue = $this->inventoryDataHelper->getBackOrderValue();
        $useConfigBackOrder = $stockData['use_config_backorders'] ?? false;
        $backorderQty = $stockData['backorders'] ?? 0;
        $thresholdQty = $stockData['min_qty'];
        if ((($useConfigBackOrder && !$backOrderValue) || $backorderQty == '0')) {
            if (($qtyAfter - $thresholdQty) <= 0) {
                $stockStatus = 0;
            } else {
                $stockStatus = 1;
            }
        } elseif ($qtyChange <= 0) {
            $stockStatus = 0;
        } else {
            $stockStatus = 1;
        }
        return $stockStatus;
    }

    /**
     * @param int|float|string $qtyChange
     * @return string
     */
    protected function getQtyChange($qtyChange)
    {
        if ($qtyChange > 0) {
            return $qtyChange;
        }
        $qtyChange = abs($qtyChange);
        return 0 - $qtyChange;
    }
}
