<?php
/**
 * BSS Commerce Co.
 *
 * NOTICE OF LICENSE
 *
 * @category   BSS
 * @package    Bss_InventoryReport
 * @author     Extension Team
 * @copyright  Copyright (c) 2018-2019 BSS Commerce Co. ( http://bsscommerce.com )
 * @license    http://bsscommerce.com/Bss-Commerce-License.txt
 */

namespace Bss\InventoryReport\Observer;

use Bss\InventoryReport\Helper\InventoryData as InventoryDataHelper;
use Bss\InventoryReport\Model\MultiInventorySource\MsiObjectManagement;
use Bss\InventoryReport\Model\MultiInventorySource\StockAndSourceProcessor;
use Bss\InventoryReport\Model\ReportFactory;
use Magento\Catalog\Api\Data\ProductInterface;
use Magento\Catalog\Api\ProductRepositoryInterface;
use Magento\Catalog\Model\Product;
use Magento\Framework\Event\Observer;
use Magento\Framework\Event\ObserverInterface;
use Magento\Framework\Exception\InputException;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Framework\Stdlib\DateTime\TimezoneInterface;
use Magento\Sales\Model\Order\Shipment;
use Magento\Store\Model\StoreManagerInterface;

class SourceDeductionProcessor implements ObserverInterface
{
    /**
     * @var StoreManagerInterface
     */
    protected $storeManagerInterface;

    /**
     * @var MsiObjectManagement
     */
    protected $msiObjectManagement;

    /**
     * @var StockAndSourceProcessor
     */
    protected $stockAndSourceProcessor;

    /**
     * @var TimezoneInterface
     */
    protected $timezone;

    /**
     * @var ProductRepositoryInterface
     */
    protected $productRepository;

    /**
     * @var ReportFactory
     */
    protected $reportFactory;

    /**
     * @var InventoryDataHelper
     */
    protected $inventoryDataHelper;

    /**
     * SourceDeductionProcessor constructor.
     * @param StoreManagerInterface $storeManagerInterface
     * @param MsiObjectManagement $msiObjectManagement
     * @param StockAndSourceProcessor $stockAndSourceProcessor
     * @param TimezoneInterface $timezone
     * @param ProductRepositoryInterface $productRepository
     * @param ReportFactory $reportFactory
     * @param InventoryDataHelper $inventoryDataHelper
     */
    public function __construct(
        StoreManagerInterface $storeManagerInterface,
        MsiObjectManagement $msiObjectManagement,
        StockAndSourceProcessor $stockAndSourceProcessor,
        TimezoneInterface $timezone,
        ProductRepositoryInterface $productRepository,
        ReportFactory $reportFactory,
        InventoryDataHelper $inventoryDataHelper
    ) {
        $this->storeManagerInterface = $storeManagerInterface;
        $this->msiObjectManagement = $msiObjectManagement;
        $this->stockAndSourceProcessor = $stockAndSourceProcessor;
        $this->timezone = $timezone;
        $this->productRepository = $productRepository;
        $this->reportFactory = $reportFactory;
        $this->inventoryDataHelper = $inventoryDataHelper;
    }

    /**
     * @param Observer $observer
     * @throws InputException
     * @throws LocalizedException
     * @throws NoSuchEntityException
     */
    public function execute(Observer $observer)
    {
        // Only process this if MSI is enabled
        if ($this->msiObjectManagement->isMsiEnabled()) {
            /** @var Shipment $shipment */
            $shipment = $observer->getEvent()->getShipment();
            if ($shipment->getOrigData('entity_id')) {
                return;
            }

            $data = [];
            $data['time'] = $this->timezone->date();
            $data['user'] = $this->inventoryDataHelper->getAdminName();
            $data['note'] = $this->getNote($shipment->getIncrementId());

            $sourceCode = $this->getSourceCode($shipment);
            if ($sourceCode) {
                $data['source_code'] = $sourceCode;
                $stockId = $this->getStockId($shipment);
                $shipmentItems = $this->msiObjectManagement->getItemsToDeductFromShipment()->execute($shipment);
                foreach ($shipmentItems as $item) {
                    $qtyChange = $item->getQty();
                    $productSku = $item->getSku();

                    $product = $this->productRepository->get($productSku);
                    $sourceItem = $this->stockAndSourceProcessor->getSourceItem($product->getSku(), $sourceCode);
                    $stockQtyAfter = $sourceItem->getQuantity();

                    $data['product_id'] = $product->getId();
                    $data['product_name'] = $product->getName();
                    $data['product_sku'] = $product->getSku();
                    $data['qty_after'] = $stockQtyAfter;
                    $data['qty_change'] = 0 - $qtyChange;
                    $data['saleable_qty'] = $this->stockAndSourceProcessor->getSaleableQtyByStockId(
                        $productSku,
                        $stockId
                    );
                    $data['saleable_qty_change'] = null;
                    $data['stock_status'] = $this->getStockStatus($product, $stockQtyAfter, $sourceItem->getStatus());
                    $data['stock_id'] = $stockId;
                    $data['shipment_entity_id'] = $shipment->getId();
                    $this->reportFactory->create()->setData($data)->save();
                }
            }
        }
    }

    /**
     * @param int|string $shipmentId
     * @return string
     */
    protected function getNote($shipmentId)
    {
        return __('New Shipment No: %1', $shipmentId);
    }

    /**
     * @param Shipment $shipment
     * @return string|null
     */
    protected function getSourceCode($shipment)
    {
        if (!empty($shipment->getExtensionAttributes())
            && !empty($shipment->getExtensionAttributes()->getSourceCode())) {
            return $shipment->getExtensionAttributes()->getSourceCode();
        } elseif ($this->msiObjectManagement->getIsSingleSourceMode()->execute()) {
            return $this->msiObjectManagement->getDefaultSourceProvider()->getCode();
        }
        return null;
    }

    /**
     * @param Shipment $shipment
     * @return int|string
     * @throws NoSuchEntityException
     */
    protected function getStockId($shipment)
    {
        $store = $this->storeManagerInterface->getStore($shipment->getStoreId());
        $websiteCode = $store->getWebsite()->getCode();
        $stock = $this->stockAndSourceProcessor->getStockByWebsiteCode($websiteCode);
        return $stock->getStockId();
    }

    /**
     * @param Product|ProductInterface $product
     * @param $qtyAfter
     * @param $status
     * @return int
     */
    public function getStockStatus($product, $qtyAfter, $status)
    {
        if ($status && (float)$qtyAfter <= 0) {
            if ($product->getExtensionAttributes()->getStockItem()->getBackorders() == 0) {
                return 0;
            }
        }
        return $status;
    }
}
