<?php
/**
 * BSS Commerce Co.
 *
 * NOTICE OF LICENSE
 *
 * @category   BSS
 * @package    Bss_InventoryReport
 * @author     Extension Team
 * @copyright  Copyright (c) 2018-2019 BSS Commerce Co. ( http://bsscommerce.com )
 * @license    http://bsscommerce.com/Bss-Commerce-License.txt
 */

namespace Bss\InventoryReport\Plugin\Attribute;

use Bss\InventoryReport\Helper\InventoryData as InventoryDataHelper;
use Bss\InventoryReport\Model\MultiInventorySource\MsiObjectManagement;
use Bss\InventoryReport\Model\MultiInventorySource\StockAndSourceProcessor;
use Bss\InventoryReport\Model\ReportFactory;
use Bss\InventoryReport\Model\ResourceModel\Report as InventoryResourceReport;
use Exception;
use Magento\Backend\Model\Session;
use Magento\Catalog\Api\Data\ProductInterface;
use Magento\Catalog\Api\ProductRepositoryInterface;
use Magento\Catalog\Controller\Adminhtml\Product\Action\Attribute\Save as UpdateAttribute;
use Magento\Catalog\Helper\Product\Edit\Action\Attribute;
use Magento\CatalogInventory\Api\StockConfigurationInterface;
use Magento\CatalogInventory\Api\StockRegistryInterface;
use Magento\Framework\App\RequestInterface;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\Message\ManagerInterface;
use Magento\Framework\Stdlib\DateTime\DateTime;
use Magento\Framework\Stdlib\DateTime\TimezoneInterface;

class Save
{
    const DEFAULT_SOURCE = 'default';

    /**
     * @var StockRegistryInterface
     */
    protected $stockRegistry;

    /**
     * @var Attribute
     */
    protected $attributeHelper;

    /**
     * @var StockConfigurationInterface
     */
    protected $stockConfiguration;

    /**
     * @var ManagerInterface
     */
    protected $messageManager;

    /**
     * @var Session
     */
    protected $backendSession;

    /**
     * @var MsiObjectManagement
     */
    protected $msiObjectManagement;

    /**
     * @var StockAndSourceProcessor
     */
    protected $stockAndSourceProcessor;

    /**
     * @var ProductRepositoryInterface
     */
    protected $productRepository;

    /**
     * @var TimezoneInterface
     */
    protected $timezone;

    /**
     * @var ReportFactory
     */
    protected $reportFactory;

    /**
     * @var DateTime
     */
    protected $dateTime;

    /**
     * @var InventoryResourceReport
     */
    protected $inventoryResourceReport;

    /**
     * @var InventoryDataHelper
     */
    protected $inventoryDataHelper;

    /**
     * Save constructor.
     * @param StockRegistryInterface $stockRegistry
     * @param Attribute $attributeHelper
     * @param StockConfigurationInterface $stockConfiguration
     * @param ManagerInterface $messageManager
     * @param Session $backendSession
     * @param MsiObjectManagement $msiObjectManagement
     * @param StockAndSourceProcessor $stockAndSourceProcessor
     * @param ProductRepositoryInterface $productRepository
     * @param TimezoneInterface $timezone
     * @param ReportFactory $reportFactory
     * @param DateTime $dateTime
     * @param InventoryResourceReport $inventoryResourceReport
     * @param InventoryDataHelper $inventoryDataHelper
     */
    public function __construct(
        StockRegistryInterface $stockRegistry,
        Attribute $attributeHelper,
        StockConfigurationInterface $stockConfiguration,
        ManagerInterface $messageManager,
        Session $backendSession,
        MsiObjectManagement $msiObjectManagement,
        StockAndSourceProcessor $stockAndSourceProcessor,
        ProductRepositoryInterface $productRepository,
        TimezoneInterface $timezone,
        ReportFactory $reportFactory,
        DateTime $dateTime,
        InventoryResourceReport $inventoryResourceReport,
        InventoryDataHelper $inventoryDataHelper
    ) {
        $this->stockRegistry = $stockRegistry;
        $this->attributeHelper = $attributeHelper;
        $this->stockConfiguration = $stockConfiguration;
        $this->messageManager = $messageManager;
        $this->backendSession = $backendSession;
        $this->msiObjectManagement = $msiObjectManagement;
        $this->stockAndSourceProcessor = $stockAndSourceProcessor;
        $this->productRepository = $productRepository;
        $this->timezone = $timezone;
        $this->reportFactory = $reportFactory;
        $this->dateTime = $dateTime;
        $this->inventoryResourceReport = $inventoryResourceReport;
        $this->inventoryDataHelper = $inventoryDataHelper;
    }

    /**
     * Report data to grid
     *
     * @param UpdateAttribute $subject
     * @param \Closure $proceed
     * @return void
     */
    public function aroundExecute(
        UpdateAttribute $subject,
        \Closure $proceed
    ) {
        /** @var RequestInterface $request */
        $request = $subject->getRequest();
        if ($this->msiObjectManagement->isMsiEnabled()) {
            $this->processWithMsi($request);
            return $proceed();
        }
        $this->processWithoutMsi($request);
        return $proceed();
    }

    /**
     * @param RequestInterface $request
     */
    protected function processWithMsi($request)
    {
        try {
            $inventoryData = $request->getParam('inventory', []);
            $inventoryData = $this->addConfigSettings($inventoryData);
            $dataQtyBefore = $this->backendSession->getBssStorageQtyBefore();

            $storeId = $this->attributeHelper->getSelectedStoreId();
            $websiteId = $this->attributeHelper->getStoreWebsiteId($storeId);
            $productIds = $this->attributeHelper->getProductIds();

            if (!empty($inventoryData)) {
                $this->updateInventoryReport($dataQtyBefore, $productIds, $websiteId, $inventoryData);
            }
        } catch (Exception $e) {
            $this->messageManager->addExceptionMessage(
                $e,
                __('Something went wrong while updating the product(s) attributes.')
            );
        }
    }

    /**
     * Get Default Config
     *
     * @param array $inventoryData
     * @return array
     */
    protected function addConfigSettings($inventoryData)
    {
        $options = $this->stockConfiguration->getConfigItemOptions();
        foreach ($options as $option) {
            $useConfig = 'use_config_' . $option;
            if (isset($inventoryData[$option]) && !isset($inventoryData[$useConfig])) {
                $inventoryData[$useConfig] = 0;
            }
        }
        return $inventoryData;
    }

    /**
     * Update data to report grid
     *
     * @param array $dataQtyBefore
     * @param array $productIds
     * @param int $websiteId
     * @param int $inventoryData
     * @throws LocalizedException
     */
    protected function updateInventoryReport($dataQtyBefore, $productIds, $websiteId, $inventoryData)
    {
        foreach ($productIds as $productId) {
            $product = $this->productRepository->getById($productId);
            $productSku = $product->getSku();
            $productName = $product->getName();

            $data = [];
            $data['time'] = $this->timezone->date();
            $data['user'] = $this->inventoryDataHelper->getAdminName();
            $data['note'] = __('Admin Update Attributes');
            $data['source_code'] = self::DEFAULT_SOURCE;
            $data['product_id'] = $productId;
            $data['product_name'] = $productName;
            $data['product_sku'] = $productSku;

            $stockItemDo = $this->stockRegistry->getStockItem($productId, $websiteId);
            $backOrderConfig = $stockItemDo->getBackorders() == '0';
            $stockStatus = $stockItemDo->getIsInStock() ? 1 : 0;
            $qtyAfter = $stockItemDo->getQty();
            $threshSoldBefore = $stockItemDo->getMinQty();

            $useConfigBackOrder = $stockItemDo->getData('use_config_backorders');
            $backOrderValue = $this->inventoryDataHelper->getBackOrderValue();

            if (((float)$inventoryData['qty'] <= (float)$threshSoldBefore) &&
                ($backOrderConfig || ($useConfigBackOrder && $backOrderValue == '0'))) {
                $stockStatus = 0;
            }

            $sourceStocks = $this->stockAndSourceProcessor->getStockBySourceCode(self::DEFAULT_SOURCE);
            $qtySaleAfter = $stockId = null;
            foreach ($sourceStocks as $sourceStock) {
                $stockId = $sourceStock->getStockId();
                $qtySaleAfter = $this->stockAndSourceProcessor->getSaleableQtyByStockId($productSku, $stockId);
            }
            if (!empty($inventoryData)) {
                $qtyBefore = $dataQtyBefore[$productId];
                $qtyChange = isset($inventoryData['qty']) ? $inventoryData['qty'] - $qtyBefore : 0;
                $threshSoldAfter = (float) ($inventoryData['min_qty'] ?? $threshSoldBefore);
                $salableQtyChange = $qtyChange - ($threshSoldAfter - $threshSoldBefore);
                $data['qty_change'] = $qtyChange;
                $data['stock_status'] = $stockStatus;
                $data['qty_after'] = $inventoryData['qty'] ?? $qtyAfter;
                $data['saleable_qty_change'] = $salableQtyChange > 0 ? $salableQtyChange : '-' . abs($salableQtyChange);
                $data['saleable_qty'] = $qtySaleAfter + $salableQtyChange;
                $data['stock_id'] = $stockId;
                $this->reportFactory->create()->setData($data)->save();
            }
        }
    }

    /**
     * @param RequestInterface $request
     */
    protected function processWithoutMsi($request)
    {
        $inventoryData = $request->getParam('inventory', []);
        $currentTime = $this->dateTime->gmtDate();
        $adminName = $this->inventoryDataHelper->getAdminName();
        $backOrderValue = $this->inventoryDataHelper->getBackOrderValue();

        if (is_array($inventoryData) &&
            !empty($inventoryData) &&
            isset($inventoryData['qty'])) {
            $productIds = $this->attributeHelper->getProductIds();
            foreach ($productIds as $productId) {
                try {
                    $product = $this->productRepository->getById($productId);
                    if ($this->isAllowProductType($product)) {
                        $productName = $product->getName();
                        $productSku = $product->getSku();

                        $stockItem = $this->stockRegistry->getStockItem($productId);
                        $stockQty = $stockItem->getQty();
                        $minQty = $stockItem->getMinQty();
                        $backOrderConfigProduct = $stockItem->getBackorders();
                        $useConfigBackOrder = $stockItem->getData('use_config_backorders');

                        $stockQtyAfter = $inventoryData['qty'];
                        $qtyChange = $this->getQtyChange($stockQty, $stockQtyAfter);

                        $stockStatus = $stockItem->getIsInStock();
                        $stockStatus = $this->checkStatus(
                            $backOrderValue,
                            $backOrderConfigProduct,
                            $stockStatus,
                            $stockQtyAfter,
                            $minQty,
                            $useConfigBackOrder,
                            $request
                        );

                        $note = __('Admin Update Attributes');
                        if ($qtyChange) {
                            $this->inventoryResourceReport->saveReport(
                                $currentTime,
                                $productId,
                                $productName,
                                $productSku,
                                $stockQtyAfter,
                                $qtyChange,
                                $stockStatus,
                                $adminName,
                                $note,
                                null,
                                null,
                                null
                            );
                        }
                    }
                } catch (Exception $e) {
                    $this->messageManager->addExceptionMessage(
                        $e,
                        __('Something went wrong while updating the product(s) attributes.')
                    );
                }
            }
        }
    }

    /**
     * @param ProductInterface $product
     * @return bool
     */
    protected function isAllowProductType($product)
    {
        $productType = $product->getTypeId();
        return $productType == 'simple' ||
            $productType == 'downloadable' ||
            $productType == 'virtual';
    }

    /**
     * @param string|int|float $stockQty
     * @param string|int|float $stockQtyAfter
     * @return string
     */
    protected function getQtyChange($stockQty, $stockQtyAfter)
    {
        $qtyChange = $stockQtyAfter - $stockQty;
        if ($qtyChange > 0) {
            return $qtyChange;
        }
        $qtyChange = abs($qtyChange);
        return 0 - $qtyChange;
    }

    /**
     * @param string $backOrderValue
     * @param string $backOrderConfigProduct
     * @param string $stockStatus
     * @param string $stockQtyAfter
     * @param string $minQty
     * @param string $useConfigBackOrder
     * @param string $request
     * @return string
     */
    protected function checkStatus(
        $backOrderValue,
        $backOrderConfigProduct,
        $stockStatus,
        $stockQtyAfter,
        $minQty,
        $useConfigBackOrder,
        $request
    ) {
        $inventoryData = $request->getParam('inventory', []);
        if ((float)$stockQtyAfter <= (float)$minQty &&
            ($backOrderConfigProduct == '0' || ($useConfigBackOrder && $backOrderValue == '0'))) {
            $stockStatus = 0;
        } else {
            if (isset($inventoryData['is_in_stock'])) {
                $stockStatus = 0;
                if ($inventoryData['is_in_stock'] == 1) {
                    $stockStatus = 1;
                }
            } elseif (!$stockStatus || !$stockQtyAfter) {
                $stockStatus = 0;
            } else {
                $stockStatus = 0;
                if ($stockQtyAfter <= $minQty) {
                    $stockStatus = 0;
                }
            }
        }
        return $stockStatus;
    }
}
