<?php
/**
 * BSS Commerce Co.
 *
 * NOTICE OF LICENSE
 *
 * @category   BSS
 * @package    Bss_InventoryReport
 * @author     Extension Team
 * @copyright  Copyright (c) 2018-2019 BSS Commerce Co. ( http://bsscommerce.com )
 * @license    http://bsscommerce.com/Bss-Commerce-License.txt
 */

namespace Bss\InventoryReport\Plugin\Model\Credit;

use Bss\InventoryReport\Helper\InventoryData as InventoryDataHelper;
use Bss\InventoryReport\Model\MultiInventorySource\MsiObjectManagement;
use Bss\InventoryReport\Model\MultiInventorySource\StockAndSourceProcessor;
use Bss\InventoryReport\Model\ReportFactory;
use Magento\Catalog\Api\ProductRepositoryInterface;
use Magento\Framework\Exception\InputException;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Framework\Registry;
use Magento\Framework\Stdlib\DateTime\TimezoneInterface;
use Magento\Sales\Api\Data\OrderInterface;
use Magento\Sales\Model\Order;
use Magento\Store\Model\StoreManagerInterface;

class ProcessRefundItems
{
    /**
     * @var Registry
     */
    protected $registry;

    /**
     * @var ReportFactory
     */
    protected $reportFactory;

    /**
     * @var StoreManagerInterface
     */
    protected $storeManager;

    /**
     * @var MsiObjectManagement
     */
    protected $msiObjectManagement;

    /**
     * @var StockAndSourceProcessor
     */
    protected $stockAndSourceProcessor;

    /**
     * @var ProductRepositoryInterface
     */
    protected $productRepository;

    /**
     * @var TimezoneInterface
     */
    protected $timezone;

    /**
     * @var InventoryDataHelper
     */
    protected $inventoryDataHelper;

    /**
     * ProcessRefundItems constructor.
     * @param Registry $registry
     * @param ReportFactory $reportFactory
     * @param StoreManagerInterface $storeManager
     * @param MsiObjectManagement $msiObjectManagement
     * @param StockAndSourceProcessor $stockAndSourceProcessor
     * @param ProductRepositoryInterface $productRepository
     * @param TimezoneInterface $timezone
     * @param InventoryDataHelper $inventoryDataHelper
     */
    public function __construct(
        Registry $registry,
        ReportFactory $reportFactory,
        StoreManagerInterface $storeManager,
        MsiObjectManagement $msiObjectManagement,
        StockAndSourceProcessor $stockAndSourceProcessor,
        ProductRepositoryInterface $productRepository,
        TimezoneInterface $timezone,
        InventoryDataHelper $inventoryDataHelper
    ) {
        $this->registry = $registry;
        $this->reportFactory = $reportFactory;
        $this->storeManager = $storeManager;
        $this->msiObjectManagement = $msiObjectManagement;
        $this->stockAndSourceProcessor = $stockAndSourceProcessor;
        $this->productRepository = $productRepository;
        $this->timezone = $timezone;
        $this->inventoryDataHelper = $inventoryDataHelper;
    }

    /**
     * @param \Magento\InventorySales\Model\ReturnProcessor\ProcessRefundItems $subject
     * @param callable $proceed
     * @param OrderInterface $order
     * @param array $itemsToRefund
     * @param array $returnToStockItems
     * @return mixed
     * @throws NoSuchEntityException
     * @throws InputException
     * @throws LocalizedException
     * @SuppressWarnings(PHPMD.UnusedFormalParameter)
     */
    public function aroundExecute(
        \Magento\InventorySales\Model\ReturnProcessor\ProcessRefundItems $subject,
        callable $proceed,
        OrderInterface $order,
        array $itemsToRefund,
        array $returnToStockItems
    ) {
        // Only process this if MSI is enabled
        if ($this->msiObjectManagement->isMsiEnabled()) {
            $creditId = $this->registry->registry('bss_creadit_memo_id');
            $inCrementId = $this->registry->registry('bss_creadit_memo_incrementid');
            $deductedItems = $this->msiObjectManagement->getSourceDeductedOrderItems()->execute(
                $order,
                $returnToStockItems
            );

            $websiteCode = $this->getWebsiteCode($order);
            $stock = $this->stockAndSourceProcessor->getStockByWebsiteCode($websiteCode);
            $stockId = $stock->getStockId();

            foreach ($itemsToRefund as $item) {
                $sku = $item->getSku();
                $totalDeductedQty = $this->getTotalDeductedQty($item, $deductedItems);
                $processedQty = $item->getProcessedQuantity() - $totalDeductedQty;
                $qtyBackToSource = ($processedQty > 0) ? $item->getQuantity() - $processedQty : $item->getQuantity();

                if ($qtyBackToSource > 0) {
                    foreach ($deductedItems as $deductedItemResult) {
                        $data = [];
                        $product = $this->productRepository->get($sku);
                        $sourceCode = $deductedItemResult->getSourceCode();
                        $sourceItem = $this->stockAndSourceProcessor->getSourceItem($sku, $sourceCode);
                        if ($sourceItem) {
                            $salableQty = $this->stockAndSourceProcessor->getSaleableQtyByStockId(
                                $sku,
                                $stockId
                            );

                            $data['time'] = $this->timezone->date();
                            $data['user'] = $this->inventoryDataHelper->getAdminName();
                            $data['note'] = __('New Credit No: %1', $inCrementId);
                            $data['source_code'] = $sourceCode;
                            $data['product_id'] = $product->getId();
                            $data['product_name'] = $product->getName();
                            $data['product_sku'] = $sku;
                            $data['qty_after'] = $sourceItem->getQuantity() + $qtyBackToSource;
                            $data['qty_change'] = $qtyBackToSource;
                            $data['stock_status'] = $sourceItem->getStatus();
                            $data['saleable_qty_change'] = $qtyBackToSource;
                            $data['saleable_qty'] = $salableQty + $qtyBackToSource;
                            $data['stock_id'] = $stockId;
                            $data['credit_entity_id'] = $creditId;
                            $this->reportFactory->create()->setData($data)->save();
                        }
                    }
                }
            }
        }
        return $proceed($order, $itemsToRefund, $returnToStockItems);
    }

    /**
     * @param Order $order
     * @return string
     * @throws NoSuchEntityException
     */
    protected function getWebsiteCode($order)
    {
        $store = $this->storeManager->getStore($order->getStoreId());
        return $store->getWebsite()->getCode();
    }

    /**
     * @param $item
     * @param array $deductedItems
     * @return float
     */
    private function getTotalDeductedQty($item, array $deductedItems): float
    {
        $result = 0;

        foreach ($deductedItems as $deductedItemResult) {
            foreach ($deductedItemResult->getItems() as $deductedItem) {
                if ($item->getSku() != $deductedItem->getSku()) {
                    continue;
                }
                $result += $deductedItem->getQuantity();
            }
        }

        return $result;
    }
}
