<?php
/**
 * BSS Commerce Co.
 *
 * NOTICE OF LICENSE
 *
 * @category   BSS
 * @package    Bss_InventoryReport
 * @author     Extension Team
 * @copyright  Copyright (c) 2018-2019 BSS Commerce Co. ( http://bsscommerce.com )
 * @license    http://bsscommerce.com/Bss-Commerce-License.txt
 */
namespace Bss\InventoryReport\Plugin\Model\Import;

use Bss\InventoryReport\Model\ReportFactory;
use Exception;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\ImportExport\Model\ResourceModel\Import\Data as ImportData;
use Magento\InventoryImportExport\Model\Import\Sources as SourceImportModel;
use Psr\Log\LoggerInterface;
use Magento\Framework\Api\SearchCriteriaBuilder;
use Magento\Catalog\Model\ProductRepository;
use Bss\InventoryReport\Model\MultiInventorySource\MsiObjectManagement;
use Bss\InventoryReport\Model\MultiInventorySource\StockAndSourceProcessor;
use Magento\Framework\Stdlib\DateTime\TimezoneInterface;
use Bss\InventoryReport\Helper\InventoryData as InventoryDataHelper;

class Sources
{
    /**
     * @var ImportData
     */
    protected $dataSourceModel;

    /**
     * @var ReportFactory
     */
    protected $reportFactory;

    /**
     * @var LoggerInterface
     */
    protected $logger;

    /**
     * @var SearchCriteriaBuilder
     */
    protected $searchCriteriaBuilder;

    /**
     * @var ProductRepository
     */
    protected $productRepository;

    /**
     * @var MsiObjectManagement
     */
    protected $msiObjectManagement;

    /**
     * @var StockAndSourceProcessor
     */
    protected $stockAndSourceProcessor;

    /**
     * @var TimezoneInterface
     */
    protected $timezone;

    /**
     * @var InventoryDataHelper
     */
    protected $inventoryDataHelper;

    /**
     * Sources constructor.
     * @param ImportData $importData
     * @param ReportFactory $reportFactory
     * @param LoggerInterface $logger
     * @param SearchCriteriaBuilder $searchCriteriaBuilder
     * @param ProductRepository $productRepository
     * @param MsiObjectManagement $msiObjectManagement
     * @param StockAndSourceProcessor $stockAndSourceProcessor
     * @param TimezoneInterface $timezone
     * @param InventoryDataHelper $inventoryDataHelper
     */
    public function __construct(
        ImportData $importData,
        ReportFactory $reportFactory,
        LoggerInterface $logger,
        SearchCriteriaBuilder $searchCriteriaBuilder,
        ProductRepository $productRepository,
        MsiObjectManagement $msiObjectManagement,
        StockAndSourceProcessor $stockAndSourceProcessor,
        TimezoneInterface $timezone,
        InventoryDataHelper $inventoryDataHelper
    ) {
        $this->dataSourceModel = $importData;
        $this->reportFactory = $reportFactory;
        $this->logger = $logger;
        $this->searchCriteriaBuilder = $searchCriteriaBuilder;
        $this->productRepository = $productRepository;
        $this->msiObjectManagement = $msiObjectManagement;
        $this->stockAndSourceProcessor = $stockAndSourceProcessor;
        $this->timezone = $timezone;
        $this->inventoryDataHelper = $inventoryDataHelper;
    }

    /**
     * Save Report When Import Data
     *
     * @param SourceImportModel $subject
     * @param \Closure $proceed
     * @return mixed
     * @throws NoSuchEntityException
     * @throws Exception
     * @SuppressWarnings(PHPMD.UnusedFormalParameter)
     */
    public function aroundImportData(
        SourceImportModel $subject,
        \Closure $proceed
    ) {
        // Only process if MSI is enabled
        if ($this->msiObjectManagement->isMsiEnabled()) {
            while ($bunch = $this->dataSourceModel->getNextBunch()) {
                $this->eachData($bunch);
            }
        }
        return $proceed();
    }

    /**
     * @param array $bunch
     */
    protected function eachData($bunch)
    {
        $adminUser = $this->inventoryDataHelper->getAdminName();
        foreach ($bunch as $rowData) {
            $data = [];
            $data['time'] = $this->timezone->date();
            $data['user'] = $adminUser;
            $data['note'] = __('Import Product');
            $sourceCode = $rowData[SourceImportModel::COL_SOURCE_CODE];
            $productSku = $rowData[SourceImportModel::COL_SKU];
            //---Search if SKU exist---//
            $searchCriteria = $this->searchCriteriaBuilder->addFilter(
                'sku',
                $productSku,
                'eq'
            )->create();
            $products = $this->productRepository->getList($searchCriteria);
            $Items = $products->getItems();
            //---Skip if not exist---//
            if (count($Items) === 0) {
                continue;
            }
            //---End check SKU exist---//
            $this->createRecordInventory($productSku, $sourceCode, $rowData);
        }
    }

    /**
     * @param string $productSku
     * @param string $sourceCode
     * @param array $rowData
     */
    protected function createRecordInventory($productSku, $sourceCode, $rowData)
    {
        $stockQtyAfter = 0;
        $stockQtyBefore = 0;
        $productStatus = false;
        $product = null;
        try {
            $sourceItem = $this->stockAndSourceProcessor->getSourceItem($productSku, $sourceCode);
            $stockQtyBefore = $sourceItem ? (int)$sourceItem->getQuantity() : 0;
            $stockQtyAfter = (int)($rowData[SourceImportModel::COL_QTY]);
            $productStatus = $sourceItem ? (int)$sourceItem->getStatus() : 0;
            $product = $this->productRepository->get($productSku);
        } catch (Exception $e) {
            $this->logger->critical(__('Sku not exist: ') . $productSku . $e->getMessage());
        }

        if (isset($rowData[SourceImportModel::COL_STATUS])) {
            $productStatus = (int)$rowData[SourceImportModel::COL_STATUS];
        }
        $qtyChange = $stockQtyAfter - $stockQtyBefore;
        if ($qtyChange && $product !== null) {
            try {
                $data['product_id'] = $product->getId();
                $data['product_name'] = $product->getName();
                $data['product_sku'] = $productSku;
                $data['qty_after'] = $stockQtyAfter;
                $data['qty_change'] = $qtyChange;
                $data['stock_status'] = $productStatus;
                $data['source_code'] = $sourceCode;
                $this->reportFactory->create()->setData($data)->save();
            } catch (Exception $e) {
                $this->logger->debug($e->getMessage());
            }
        }
    }
}
