<?php
/**
 * BSS Commerce Co.
 *
 * NOTICE OF LICENSE
 *
 * @category   BSS
 * @package    Bss_InventoryReport
 * @author     Extension Team
 * @copyright  Copyright (c) 2018-2019 BSS Commerce Co. ( http://bsscommerce.com )
 * @license    http://bsscommerce.com/Bss-Commerce-License.txt
 */

namespace Bss\InventoryReport\Plugin\Model;

use Bss\InventoryReport\Helper\InventoryData as InventoryDataHelper;
use Bss\InventoryReport\Model\MultiInventorySource\MsiObjectManagement;
use Bss\InventoryReport\Model\MultiInventorySource\StockAndSourceProcessor;
use Bss\InventoryReport\Model\ReportFactory;
use Magento\Catalog\Api\ProductRepositoryInterface;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Framework\Stdlib\DateTime\TimezoneInterface;
use Magento\InventoryApi\Api\Data\SourceItemSearchResultsInterface;
use Magento\InventoryCatalog\Model\ResourceModel\BulkInventoryTransfer;

class ProcessTransfer
{
    /**
     * @var ReportFactory
     */
    protected $reportFactory;

    /**
     * @var MsiObjectManagement
     */
    protected $msiObjectManagement;

    /**
     * @var StockAndSourceProcessor
     */
    protected $stockAndSourceProcessor;

    /**
     * @var ProductRepositoryInterface
     */
    protected $productRepository;

    /**
     * @var TimezoneInterface
     */
    protected $timezone;

    /**
     * @var InventoryDataHelper
     */
    protected $inventoryDataHelper;

    /**
     * ProcessTransfer constructor.
     * @param ReportFactory $reportFactory
     * @param MsiObjectManagement $msiObjectManagement
     * @param StockAndSourceProcessor $stockAndSourceProcessor
     * @param ProductRepositoryInterface $productRepository
     * @param TimezoneInterface $timezone
     * @param InventoryDataHelper $inventoryDataHelper
     */
    public function __construct(
        ReportFactory $reportFactory,
        MsiObjectManagement $msiObjectManagement,
        StockAndSourceProcessor $stockAndSourceProcessor,
        ProductRepositoryInterface $productRepository,
        TimezoneInterface $timezone,
        InventoryDataHelper $inventoryDataHelper
    ) {
        $this->reportFactory = $reportFactory;
        $this->msiObjectManagement = $msiObjectManagement;
        $this->stockAndSourceProcessor = $stockAndSourceProcessor;
        $this->productRepository = $productRepository;
        $this->timezone = $timezone;
        $this->inventoryDataHelper = $inventoryDataHelper;
    }

    /**
     * Process save report while transfer source
     *
     * @param BulkInventoryTransfer $subject
     * @param \Closure $proceed
     * @param array $skus
     * @param string $originSource
     * @param string $destinationSource
     * @param bool $unassignFromOrigin
     * @return void
     * @throws NoSuchEntityException
     * @SuppressWarnings(PHPMD.UnusedFormalParameter)
     */
    public function aroundExecute(
        BulkInventoryTransfer $subject,
        \Closure $proceed,
        array $skus,
        string $originSource,
        string $destinationSource,
        bool $unassignFromOrigin
    ) {
        // Only process this if MSI is enabled
        if ($this->msiObjectManagement->isMsiEnabled()) {
            $types = $this->msiObjectManagement->getProductTypesBySkus()->execute($skus);
            foreach ($types as $sku => $type) {
                if ($this->msiObjectManagement->getIsSourceItemManagementAllowedForProductType()->execute($type)) {
                    $orgSourceItem = $this->stockAndSourceProcessor->getSourceItem($sku, $originSource);
                    $dstSourceItem = $this->stockAndSourceProcessor->getSourceItem($sku, $destinationSource);
                    if ($orgSourceItem && $orgSourceItem->getQuantity()) {
                        $this->processOrgSource($orgSourceItem, $sku, $originSource);
                        $this->processDstSource($orgSourceItem, $dstSourceItem, $sku, $destinationSource);
                    }
                }
            }
        }
        return $proceed($skus, $originSource, $destinationSource, $unassignFromOrigin);
    }

    /**
     * @param SourceItemSearchResultsInterface $orgSourceItem
     * @param $sku
     * @param $originSource
     * @throws NoSuchEntityException
     */
    protected function processOrgSource($orgSourceItem, $sku, $originSource)
    {
        $beforeQuantity = $orgSourceItem->getQuantity();
        $product = $this->productRepository->get($sku);
        $data['time'] = $this->timezone->date();
        $data['user'] = $this->inventoryDataHelper->getAdminName();
        $data['note'] = __('Bulk Transfer Source');
        $data['product_id'] = $product->getId();
        $data['product_name'] = $product->getName();
        $data['product_sku'] = $product->getSku();
        $data['qty_after'] = 0;
        $data['qty_change'] = 0 - $beforeQuantity;
        $data['stock_status'] = 0;
        $data['source_code'] = $originSource;
        $this->reportFactory->create()->setData($data)->save();
    }

    /**
     * @param SourceItemSearchResultsInterface $orgSourceItem
     * @param SourceItemSearchResultsInterface $dstSourceItem
     * @param string $sku
     * @param string $destinationSource
     * @throws NoSuchEntityException
     */
    protected function processDstSource($orgSourceItem, $dstSourceItem, $sku, $destinationSource)
    {
        $originSourceQty = $beforeQuantity = 0;
        $status = "";
        if ($orgSourceItem) {
            $status = $orgSourceItem->getStatus();
            $originSourceQty = $orgSourceItem->getQuantity();
        }
        if ($dstSourceItem) {
            $beforeQuantity = $dstSourceItem->getQuantity();
        }

        $product = $this->productRepository->get($sku);
        $data['time'] = $this->timezone->date();
        $data['user'] = $this->inventoryDataHelper->getAdminName();
        $data['note'] = __('Bulk Transfer Source');
        $data['product_id'] = $product->getId();
        $data['product_name'] = $product->getName();
        $data['product_sku'] = $product->getSku();
        $data['qty_after'] = $beforeQuantity + $originSourceQty;
        $data['qty_change'] = $originSourceQty;
        $data['stock_status'] = $status;
        $data['source_code'] = $destinationSource;
        $this->reportFactory->create()->setData($data)->save();
    }
}
