<?php
/**
 * BSS Commerce Co.
 *
 * NOTICE OF LICENSE
 *
 * @category   BSS
 * @package    Bss_InventoryReport
 * @author     Extension Team
 * @copyright  Copyright (c) 2018-2019 BSS Commerce Co. ( http://bsscommerce.com )
 * @license    http://bsscommerce.com/Bss-Commerce-License.txt
 */

namespace Bss\InventoryReport\Plugin\Model\SourceItem;

use Bss\InventoryReport\Helper\InventoryData as InventoryDataHelper;
use Bss\InventoryReport\Model\MultiInventorySource\MsiObjectManagement;
use Bss\InventoryReport\Model\MultiInventorySource\StockAndSourceProcessor;
use Bss\InventoryReport\Model\ReportFactory;
use Exception;
use Magento\Catalog\Api\ProductRepositoryInterface;
use Magento\Framework\App\State;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\Stdlib\DateTime\TimezoneInterface;
use Magento\Inventory\Model\ResourceModel\SourceItem\SaveMultiple as SourceSave;

class SaveMultiple
{
    /**
     * @var State
     */
    protected $state;

    /**
     * @var MsiObjectManagement
     */
    protected $msiObjectManagement;

    /**
     * @var StockAndSourceProcessor
     */
    protected $stockAndSourceProcessor;

    /**
     * @var ProductRepositoryInterface
     */
    protected $productRepository;

    /**
     * @var TimezoneInterface
     */
    protected $timezone;

    /**
     * @var ReportFactory
     */
    protected $reportFactory;

    /**
     * @var InventoryDataHelper
     */
    protected $inventoryDataHelper;

    /**
     * SaveMultiple constructor.
     * @param State $state
     * @param MsiObjectManagement $msiObjectManagement
     * @param StockAndSourceProcessor $stockAndSourceProcessor
     * @param ProductRepositoryInterface $productRepository
     * @param TimezoneInterface $timezone
     * @param ReportFactory $reportFactory
     * @param InventoryDataHelper $inventoryDataHelper
     */
    public function __construct(
        State $state,
        MsiObjectManagement $msiObjectManagement,
        StockAndSourceProcessor $stockAndSourceProcessor,
        ProductRepositoryInterface $productRepository,
        TimezoneInterface $timezone,
        ReportFactory $reportFactory,
        InventoryDataHelper $inventoryDataHelper
    ) {
        $this->state = $state;
        $this->msiObjectManagement = $msiObjectManagement;
        $this->stockAndSourceProcessor = $stockAndSourceProcessor;
        $this->productRepository = $productRepository;
        $this->timezone = $timezone;
        $this->reportFactory = $reportFactory;
        $this->inventoryDataHelper = $inventoryDataHelper;
    }

    /**
     * Save Report to Database
     *
     * @param SourceSave $subject
     * @param \Closure $proceed
     * @param array $sourceItems
     * @return mixed
     * @throws Exception
     * @SuppressWarnings(PHPMD.UnusedFormalParameter)
     */
    public function aroundExecute(
        SourceSave $subject,
        \Closure $proceed,
        array $sourceItems
    ) {
        // Only process this if MSI enabled
        if ($this->msiObjectManagement->isMsiEnabled()) {
            if ($this->isAllowAreaCode()) {
                foreach ($sourceItems as $sourceItem) {
                    $data = [];
                    $product = $this->productRepository->get($sourceItem->getSku());
                    /** Request Source Information **/
                    $sourceCode = $sourceItem->getSourceCode();
                    $stockQtyAfter = $qtySaleAfter = (int)$sourceItem->getQuantity();
                    $productStatus = $sourceItem->getStatus();

                    /** Current Source Information **/
                    $beforeSourceItem = $this->stockAndSourceProcessor->getSourceItem(
                        $product->getSku(),
                        $sourceCode
                    );
                    $stockQtyBefore = $this->getStockQtyBefore($beforeSourceItem);
                    $sourceStocks = $this->stockAndSourceProcessor->getStockBySourceCode($sourceCode);
                    foreach ($sourceStocks as $sourceStock) {
                        $stockId = $sourceStock->getStockId();
                        $qtySaleBefore = $this->stockAndSourceProcessor->getSaleableQtyByStockId(
                            $product->getSku(),
                            $stockId
                        );
                        $qtyChange = $stockQtyAfter - $stockQtyBefore;
                        $qtySaleChange = $qtySaleAfter - $qtySaleBefore;
                        if ($qtyChange || $qtySaleChange) {
                            $data['time'] = $this->timezone->date();
                            $data['user'] = $this->inventoryDataHelper->getAdminName();
                            $data['note'] = __('Rest API Source');
                            $data['product_id'] = $product->getId();
                            $data['product_name'] = $product->getName();
                            $data['product_sku'] = $product->getSku();
                            $data['qty_after'] = $stockQtyAfter;
                            $data['qty_change'] = $qtyChange;
                            $data['stock_status'] = $productStatus;
                            $data['source_code'] = $sourceCode;
                            $data['saleable_qty'] = $qtySaleAfter;
                            $data['saleable_qty_change'] = $qtySaleChange;
                            $data['stock_id'] = $stockId;
                            $this->reportFactory->create()->setData($data)->save();
                        }
                    }

                }
            }
        }
        return $proceed($sourceItems);
    }

    /**
     * @return bool
     * @throws LocalizedException
     */
    protected function isAllowAreaCode()
    {
        $areaCode = $this->state->getAreaCode();
        return $areaCode == 'webapi_rest' || $areaCode == 'webapi_soap';
    }

    /**
     * @param $beforeSourceItem
     * @return int
     */
    protected function getStockQtyBefore($beforeSourceItem)
    {
        if (!$beforeSourceItem) {
            return 0;
        }
        return (int)$beforeSourceItem->getQuantity();
    }
}
